import argparse
import typing

from kuber import management


def _populate_create(subparsers):
    """..."""
    parser = subparsers.add_parser(
        'create',
        help='Create the resource bundle.'
    )
    parser.add_argument(
        '--target', action='append',
        help=(
            'Specify the resource in the bundle to target with this '
            'operation. Only the resource specified will be acted upon. '
            'Use this flag multiple times in the same call to target '
            'multiple resources. If omitted, the default behavior is '
            'to target all resources in the bundle. Targets are specified '
            'as <NAME> or <KIND>/<NAME> or <NAMESPACE>/<KIND>/<NAME>, '
            'e.g. --target=Namespace/foo, and can contain shell-style '
            'wildcard characters for fuzzy matching.'
        )
    )
    parser.add_argument('--namespace')
    parser.add_argument(
        '--settings', action='append',
        help=(
            'Specify a file path or directory to load settings from. '
            'These settings will be available in the ResourceBundle '
            'to use in dynamically configuring the bundle.'
        )
    )


def _populate_status(subparsers):
    """..."""
    parser = subparsers.add_parser('status', help='Resource bundle status.')
    parser.add_argument('--namespace')
    parser.add_argument(
        '--target', action='append',
        help=(
            'Specify the resource in the bundle to target with this '
            'operation. Only the resource specified will be acted upon. '
            'Use this flag multiple times in the same call to target '
            'multiple resources. If omitted, the default behavior is '
            'to target all resources in the bundle. Targets are specified '
            'as <NAME> or <KIND>/<NAME> or <NAMESPACE>/<KIND>/<NAME>, '
            'e.g. --target=Namespace/foo, and can contain shell-style '
            'wildcard characters for fuzzy matching.'
        )
    )


def _populate_delete(subparsers):
    """..."""
    parser = subparsers.add_parser(
        'delete',
        help='Delete the resource bundle.'
    )
    parser.add_argument('--namespace')
    parser.add_argument(
        '--target', action='append',
        help=(
            'Specify the resource in the bundle to target with this '
            'operation. Only the resource specified will be acted upon. '
            'Use this flag multiple times in the same call to target '
            'multiple resources. If omitted, the default behavior is '
            'to target all resources in the bundle. Targets are specified '
            'as <NAME> or <KIND>/<NAME> or <NAMESPACE>/<KIND>/<NAME>, '
            'e.g. --target=Namespace/foo, and can contain shell-style '
            'wildcard characters for fuzzy matching.'
        )
    )


def _populate_render(subparsers):
    """..."""
    parser = subparsers.add_parser('render', help='Render the bundle')
    parser.add_argument('--namespace')
    parser.add_argument(
        '--settings',
        action='append',
        help=(
            'Specify a file path or directory to load settings from. '
            'These settings will be available in the ResourceBundle '
            'to use in dynamically configuring the bundle.'
        )
    )
    parser.add_argument(
        '--target', action='append',
        help=(
            'Specify the resource in the bundle to target with this '
            'operation. Only the resource specified will be acted upon. '
            'Use this flag multiple times in the same call to target '
            'multiple resources. If omitted, the default behavior is '
            'to target all resources in the bundle. Targets are specified '
            'as <NAME> or <KIND>/<NAME> or <NAMESPACE>/<KIND>/<NAME>, '
            'e.g. --target=Namespace/foo, and can contain shell-style '
            'wildcard characters for fuzzy matching.'
        )
    )


def parse_args(
        bundle: 'management.ResourceBundle',
        arguments: typing.Sequence[str] = None
) -> typing.Tuple[argparse.Namespace, typing.List[str]]:
    """Parses the CLI command line arguments."""
    parser = argparse.ArgumentParser(prog=bundle.name)
    subparsers = parser.add_subparsers(
        title='Kuber Resource Bundle Command',
        dest='command'
    )
    _populate_create(subparsers)
    _populate_delete(subparsers)
    _populate_render(subparsers)
    _populate_status(subparsers)

    return parser.parse_known_args(args=arguments)



import typing

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_13.meta_v1 import LabelSelector
from kuber.v1_13.meta_v1 import ListMeta
from kuber.v1_13.meta_v1 import ObjectMeta


class MutatingWebhookConfiguration(_kuber_definitions.Resource):
    """
    MutatingWebhookConfiguration describes the configuration of
    and admission webhook that accept or reject and may change
    the object.
    """

    def __init__(
            self,
            metadata: 'ObjectMeta' = None,
            webhooks: typing.List['Webhook'] = None,
    ):
        """Create MutatingWebhookConfiguration instance."""
        super(MutatingWebhookConfiguration, self).__init__(
            api_version='admissionregistration/v1beta1',
            kind='MutatingWebhookConfiguration'
        )
        self._properties = {
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'webhooks': webhooks if webhooks is not None else [],

        }
        self._types = {
            'apiVersion': (str, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'webhooks': (list, Webhook),

        }

    @property
    def metadata(self) -> 'ObjectMeta':
        """
        Standard object metadata; More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata.
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """
        Standard object metadata; More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata.
        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def webhooks(self) -> typing.List['Webhook']:
        """
        Webhooks is a list of webhooks and the affected resources
        and operations.
        """
        return self._properties.get('webhooks')

    @webhooks.setter
    def webhooks(
            self,
            value: typing.Union[typing.List['Webhook'], typing.List[dict]]
    ):
        """
        Webhooks is a list of webhooks and the affected resources
        and operations.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = Webhook().from_dict(item)
            cleaned.append(item)
        self._properties['webhooks'] = cleaned

    def create_resource(self, namespace: 'str' = None):
        """
        Creates the MutatingWebhookConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'create_namespaced_mutating_webhook_configuration',
            'create_mutating_webhook_configuration'
        ]

        _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )

    def replace_resource(self, namespace: 'str' = None):
        """
        Replaces the MutatingWebhookConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'replace_namespaced_mutating_webhook_configuration',
            'replace_mutating_webhook_configuration'
        ]

        _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def patch_resource(self, namespace: 'str' = None):
        """
        Patches the MutatingWebhookConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'patch_namespaced_mutating_webhook_configuration',
            'patch_mutating_webhook_configuration'
        ]

        _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def get_resource_status(self, namespace: 'str' = None):
        """This resource does not have a status."""
        pass

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the MutatingWebhookConfiguration from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_mutating_webhook_configuration',
            'read_mutating_webhook_configuration'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the MutatingWebhookConfiguration from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_mutating_webhook_configuration',
            'delete_mutating_webhook_configuration'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.AdmissionregistrationV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.AdmissionregistrationV1beta1Api(**kwargs)

    def __enter__(self) -> 'MutatingWebhookConfiguration':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class MutatingWebhookConfigurationList(_kuber_definitions.Collection):
    """
    MutatingWebhookConfigurationList is a list of
    MutatingWebhookConfiguration.
    """

    def __init__(
            self,
            items: typing.List['MutatingWebhookConfiguration'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create MutatingWebhookConfigurationList instance."""
        super(MutatingWebhookConfigurationList, self).__init__(
            api_version='admissionregistration/v1beta1',
            kind='MutatingWebhookConfigurationList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, MutatingWebhookConfiguration),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['MutatingWebhookConfiguration']:
        """
        List of MutatingWebhookConfiguration.
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['MutatingWebhookConfiguration'], typing.List[dict]]
    ):
        """
        List of MutatingWebhookConfiguration.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = MutatingWebhookConfiguration().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#types-kinds
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#types-kinds
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.AdmissionregistrationV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.AdmissionregistrationV1beta1Api(**kwargs)

    def __enter__(self) -> 'MutatingWebhookConfigurationList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class RuleWithOperations(_kuber_definitions.Definition):
    """
    RuleWithOperations is a tuple of Operations and Resources.
    It is recommended to make sure that all the tuple expansions
    are valid.
    """

    def __init__(
            self,
            api_groups: typing.List[str] = None,
            api_versions: typing.List[str] = None,
            operations: typing.List[str] = None,
            resources: typing.List[str] = None,
    ):
        """Create RuleWithOperations instance."""
        super(RuleWithOperations, self).__init__(
            api_version='admissionregistration/v1beta1',
            kind='RuleWithOperations'
        )
        self._properties = {
            'apiGroups': api_groups if api_groups is not None else [],
            'apiVersions': api_versions if api_versions is not None else [],
            'operations': operations if operations is not None else [],
            'resources': resources if resources is not None else [],

        }
        self._types = {
            'apiGroups': (list, str),
            'apiVersions': (list, str),
            'operations': (list, str),
            'resources': (list, str),

        }

    @property
    def api_groups(self) -> typing.List[str]:
        """
        APIGroups is the API groups the resources belong to. '*' is
        all groups. If '*' is present, the length of the slice must
        be one. Required.
        """
        return self._properties.get('apiGroups')

    @api_groups.setter
    def api_groups(self, value: typing.List[str]):
        """
        APIGroups is the API groups the resources belong to. '*' is
        all groups. If '*' is present, the length of the slice must
        be one. Required.
        """
        self._properties['apiGroups'] = value

    @property
    def api_versions(self) -> typing.List[str]:
        """
        APIVersions is the API versions the resources belong to. '*'
        is all versions. If '*' is present, the length of the slice
        must be one. Required.
        """
        return self._properties.get('apiVersions')

    @api_versions.setter
    def api_versions(self, value: typing.List[str]):
        """
        APIVersions is the API versions the resources belong to. '*'
        is all versions. If '*' is present, the length of the slice
        must be one. Required.
        """
        self._properties['apiVersions'] = value

    @property
    def operations(self) -> typing.List[str]:
        """
        Operations is the operations the admission hook cares about
        - CREATE, UPDATE, or * for all operations. If '*' is
        present, the length of the slice must be one. Required.
        """
        return self._properties.get('operations')

    @operations.setter
    def operations(self, value: typing.List[str]):
        """
        Operations is the operations the admission hook cares about
        - CREATE, UPDATE, or * for all operations. If '*' is
        present, the length of the slice must be one. Required.
        """
        self._properties['operations'] = value

    @property
    def resources(self) -> typing.List[str]:
        """
        Resources is a list of resources this rule applies to.

        For example: 'pods' means pods. 'pods/log' means the log
        subresource of pods. '*' means all resources, but not
        subresources. 'pods/*' means all subresources of pods.
        '*/scale' means all scale subresources. '*/*' means all
        resources and their subresources.

        If wildcard is present, the validation rule will ensure
        resources do not overlap with each other.

        Depending on the enclosing object, subresources might not be
        allowed. Required.
        """
        return self._properties.get('resources')

    @resources.setter
    def resources(self, value: typing.List[str]):
        """
        Resources is a list of resources this rule applies to.

        For example: 'pods' means pods. 'pods/log' means the log
        subresource of pods. '*' means all resources, but not
        subresources. 'pods/*' means all subresources of pods.
        '*/scale' means all scale subresources. '*/*' means all
        resources and their subresources.

        If wildcard is present, the validation rule will ensure
        resources do not overlap with each other.

        Depending on the enclosing object, subresources might not be
        allowed. Required.
        """
        self._properties['resources'] = value

    def __enter__(self) -> 'RuleWithOperations':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class ServiceReference(_kuber_definitions.Definition):
    """
    ServiceReference holds a reference to Service.legacy.k8s.io
    """

    def __init__(
            self,
            name: str = None,
            namespace: str = None,
            path: str = None,
    ):
        """Create ServiceReference instance."""
        super(ServiceReference, self).__init__(
            api_version='admissionregistration/v1beta1',
            kind='ServiceReference'
        )
        self._properties = {
            'name': name if name is not None else '',
            'namespace': namespace if namespace is not None else '',
            'path': path if path is not None else '',

        }
        self._types = {
            'name': (str, None),
            'namespace': (str, None),
            'path': (str, None),

        }

    @property
    def name(self) -> str:
        """
        `name` is the name of the service. Required
        """
        return self._properties.get('name')

    @name.setter
    def name(self, value: str):
        """
        `name` is the name of the service. Required
        """
        self._properties['name'] = value

    @property
    def namespace(self) -> str:
        """
        `namespace` is the namespace of the service. Required
        """
        return self._properties.get('namespace')

    @namespace.setter
    def namespace(self, value: str):
        """
        `namespace` is the namespace of the service. Required
        """
        self._properties['namespace'] = value

    @property
    def path(self) -> str:
        """
        `path` is an optional URL path which will be sent in any
        request to this service.
        """
        return self._properties.get('path')

    @path.setter
    def path(self, value: str):
        """
        `path` is an optional URL path which will be sent in any
        request to this service.
        """
        self._properties['path'] = value

    def __enter__(self) -> 'ServiceReference':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class ValidatingWebhookConfiguration(_kuber_definitions.Resource):
    """
    ValidatingWebhookConfiguration describes the configuration
    of and admission webhook that accept or reject and object
    without changing it.
    """

    def __init__(
            self,
            metadata: 'ObjectMeta' = None,
            webhooks: typing.List['Webhook'] = None,
    ):
        """Create ValidatingWebhookConfiguration instance."""
        super(ValidatingWebhookConfiguration, self).__init__(
            api_version='admissionregistration/v1beta1',
            kind='ValidatingWebhookConfiguration'
        )
        self._properties = {
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'webhooks': webhooks if webhooks is not None else [],

        }
        self._types = {
            'apiVersion': (str, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'webhooks': (list, Webhook),

        }

    @property
    def metadata(self) -> 'ObjectMeta':
        """
        Standard object metadata; More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata.
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """
        Standard object metadata; More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata.
        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def webhooks(self) -> typing.List['Webhook']:
        """
        Webhooks is a list of webhooks and the affected resources
        and operations.
        """
        return self._properties.get('webhooks')

    @webhooks.setter
    def webhooks(
            self,
            value: typing.Union[typing.List['Webhook'], typing.List[dict]]
    ):
        """
        Webhooks is a list of webhooks and the affected resources
        and operations.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = Webhook().from_dict(item)
            cleaned.append(item)
        self._properties['webhooks'] = cleaned

    def create_resource(self, namespace: 'str' = None):
        """
        Creates the ValidatingWebhookConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'create_namespaced_validating_webhook_configuration',
            'create_validating_webhook_configuration'
        ]

        _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )

    def replace_resource(self, namespace: 'str' = None):
        """
        Replaces the ValidatingWebhookConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'replace_namespaced_validating_webhook_configuration',
            'replace_validating_webhook_configuration'
        ]

        _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def patch_resource(self, namespace: 'str' = None):
        """
        Patches the ValidatingWebhookConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'patch_namespaced_validating_webhook_configuration',
            'patch_validating_webhook_configuration'
        ]

        _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def get_resource_status(self, namespace: 'str' = None):
        """This resource does not have a status."""
        pass

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the ValidatingWebhookConfiguration from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_validating_webhook_configuration',
            'read_validating_webhook_configuration'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the ValidatingWebhookConfiguration from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_validating_webhook_configuration',
            'delete_validating_webhook_configuration'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.AdmissionregistrationV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.AdmissionregistrationV1beta1Api(**kwargs)

    def __enter__(self) -> 'ValidatingWebhookConfiguration':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class ValidatingWebhookConfigurationList(_kuber_definitions.Collection):
    """
    ValidatingWebhookConfigurationList is a list of
    ValidatingWebhookConfiguration.
    """

    def __init__(
            self,
            items: typing.List['ValidatingWebhookConfiguration'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create ValidatingWebhookConfigurationList instance."""
        super(ValidatingWebhookConfigurationList, self).__init__(
            api_version='admissionregistration/v1beta1',
            kind='ValidatingWebhookConfigurationList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, ValidatingWebhookConfiguration),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['ValidatingWebhookConfiguration']:
        """
        List of ValidatingWebhookConfiguration.
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['ValidatingWebhookConfiguration'], typing.List[dict]]
    ):
        """
        List of ValidatingWebhookConfiguration.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = ValidatingWebhookConfiguration().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#types-kinds
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#types-kinds
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.AdmissionregistrationV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.AdmissionregistrationV1beta1Api(**kwargs)

    def __enter__(self) -> 'ValidatingWebhookConfigurationList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class Webhook(_kuber_definitions.Definition):
    """
    Webhook describes an admission webhook and the resources and
    operations it applies to.
    """

    def __init__(
            self,
            client_config: 'WebhookClientConfig' = None,
            failure_policy: str = None,
            name: str = None,
            namespace_selector: 'LabelSelector' = None,
            rules: typing.List['RuleWithOperations'] = None,
            side_effects: str = None,
    ):
        """Create Webhook instance."""
        super(Webhook, self).__init__(
            api_version='admissionregistration/v1beta1',
            kind='Webhook'
        )
        self._properties = {
            'clientConfig': client_config if client_config is not None else WebhookClientConfig(),
            'failurePolicy': failure_policy if failure_policy is not None else '',
            'name': name if name is not None else '',
            'namespaceSelector': namespace_selector if namespace_selector is not None else LabelSelector(),
            'rules': rules if rules is not None else [],
            'sideEffects': side_effects if side_effects is not None else '',

        }
        self._types = {
            'clientConfig': (WebhookClientConfig, None),
            'failurePolicy': (str, None),
            'name': (str, None),
            'namespaceSelector': (LabelSelector, None),
            'rules': (list, RuleWithOperations),
            'sideEffects': (str, None),

        }

    @property
    def client_config(self) -> 'WebhookClientConfig':
        """
        ClientConfig defines how to communicate with the hook.
        Required
        """
        return self._properties.get('clientConfig')

    @client_config.setter
    def client_config(self, value: typing.Union['WebhookClientConfig', dict]):
        """
        ClientConfig defines how to communicate with the hook.
        Required
        """
        if isinstance(value, dict):
            value = WebhookClientConfig().from_dict(value)
        self._properties['clientConfig'] = value

    @property
    def failure_policy(self) -> str:
        """
        FailurePolicy defines how unrecognized errors from the
        admission endpoint are handled - allowed values are Ignore
        or Fail. Defaults to Ignore.
        """
        return self._properties.get('failurePolicy')

    @failure_policy.setter
    def failure_policy(self, value: str):
        """
        FailurePolicy defines how unrecognized errors from the
        admission endpoint are handled - allowed values are Ignore
        or Fail. Defaults to Ignore.
        """
        self._properties['failurePolicy'] = value

    @property
    def name(self) -> str:
        """
        The name of the admission webhook. Name should be fully
        qualified, e.g., imagepolicy.kubernetes.io, where
        "imagepolicy" is the name of the webhook, and kubernetes.io
        is the name of the organization. Required.
        """
        return self._properties.get('name')

    @name.setter
    def name(self, value: str):
        """
        The name of the admission webhook. Name should be fully
        qualified, e.g., imagepolicy.kubernetes.io, where
        "imagepolicy" is the name of the webhook, and kubernetes.io
        is the name of the organization. Required.
        """
        self._properties['name'] = value

    @property
    def namespace_selector(self) -> 'LabelSelector':
        """
        NamespaceSelector decides whether to run the webhook on an
        object based on whether the namespace for that object
        matches the selector. If the object itself is a namespace,
        the matching is performed on object.metadata.labels. If the
        object is another cluster scoped resource, it never skips
        the webhook.

        For example, to run the webhook on any objects whose
        namespace is not associated with "runlevel" of "0" or "1";
        you will set the selector as follows: "namespaceSelector": {
          "matchExpressions": [
            {
              "key": "runlevel",
              "operator": "NotIn",
              "values": [
                "0",
                "1"
              ]
            }
          ]
        }

        If instead you want to only run the webhook on any objects
        whose namespace is associated with the "environment" of
        "prod" or "staging"; you will set the selector as follows:
        "namespaceSelector": {
          "matchExpressions": [
            {
              "key": "environment",
              "operator": "In",
              "values": [
                "prod",
                "staging"
              ]
            }
          ]
        }

        See https://kubernetes.io/docs/concepts/overview/working-
        with-objects/labels/ for more examples of label selectors.

        Default to the empty LabelSelector, which matches
        everything.
        """
        return self._properties.get('namespaceSelector')

    @namespace_selector.setter
    def namespace_selector(self, value: typing.Union['LabelSelector', dict]):
        """
        NamespaceSelector decides whether to run the webhook on an
        object based on whether the namespace for that object
        matches the selector. If the object itself is a namespace,
        the matching is performed on object.metadata.labels. If the
        object is another cluster scoped resource, it never skips
        the webhook.

        For example, to run the webhook on any objects whose
        namespace is not associated with "runlevel" of "0" or "1";
        you will set the selector as follows: "namespaceSelector": {
          "matchExpressions": [
            {
              "key": "runlevel",
              "operator": "NotIn",
              "values": [
                "0",
                "1"
              ]
            }
          ]
        }

        If instead you want to only run the webhook on any objects
        whose namespace is associated with the "environment" of
        "prod" or "staging"; you will set the selector as follows:
        "namespaceSelector": {
          "matchExpressions": [
            {
              "key": "environment",
              "operator": "In",
              "values": [
                "prod",
                "staging"
              ]
            }
          ]
        }

        See https://kubernetes.io/docs/concepts/overview/working-
        with-objects/labels/ for more examples of label selectors.

        Default to the empty LabelSelector, which matches
        everything.
        """
        if isinstance(value, dict):
            value = LabelSelector().from_dict(value)
        self._properties['namespaceSelector'] = value

    @property
    def rules(self) -> typing.List['RuleWithOperations']:
        """
        Rules describes what operations on what
        resources/subresources the webhook cares about. The webhook
        cares about an operation if it matches _any_ Rule. However,
        in order to prevent ValidatingAdmissionWebhooks and
        MutatingAdmissionWebhooks from putting the cluster in a
        state which cannot be recovered from without completely
        disabling the plugin, ValidatingAdmissionWebhooks and
        MutatingAdmissionWebhooks are never called on admission
        requests for ValidatingWebhookConfiguration and
        MutatingWebhookConfiguration objects.
        """
        return self._properties.get('rules')

    @rules.setter
    def rules(
            self,
            value: typing.Union[typing.List['RuleWithOperations'], typing.List[dict]]
    ):
        """
        Rules describes what operations on what
        resources/subresources the webhook cares about. The webhook
        cares about an operation if it matches _any_ Rule. However,
        in order to prevent ValidatingAdmissionWebhooks and
        MutatingAdmissionWebhooks from putting the cluster in a
        state which cannot be recovered from without completely
        disabling the plugin, ValidatingAdmissionWebhooks and
        MutatingAdmissionWebhooks are never called on admission
        requests for ValidatingWebhookConfiguration and
        MutatingWebhookConfiguration objects.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = RuleWithOperations().from_dict(item)
            cleaned.append(item)
        self._properties['rules'] = cleaned

    @property
    def side_effects(self) -> str:
        """
        SideEffects states whether this webhookk has side effects.
        Acceptable values are: Unknown, None, Some, NoneOnDryRun
        Webhooks with side effects MUST implement a reconciliation
        system, since a request may be rejected by a future step in
        the admission change and the side effects therefore need to
        be undone. Requests with the dryRun attribute will be auto-
        rejected if they match a webhook with sideEffects == Unknown
        or Some. Defaults to Unknown.
        """
        return self._properties.get('sideEffects')

    @side_effects.setter
    def side_effects(self, value: str):
        """
        SideEffects states whether this webhookk has side effects.
        Acceptable values are: Unknown, None, Some, NoneOnDryRun
        Webhooks with side effects MUST implement a reconciliation
        system, since a request may be rejected by a future step in
        the admission change and the side effects therefore need to
        be undone. Requests with the dryRun attribute will be auto-
        rejected if they match a webhook with sideEffects == Unknown
        or Some. Defaults to Unknown.
        """
        self._properties['sideEffects'] = value

    def __enter__(self) -> 'Webhook':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class WebhookClientConfig(_kuber_definitions.Definition):
    """
    WebhookClientConfig contains the information to make a TLS
    connection with the webhook
    """

    def __init__(
            self,
            ca_bundle: str = None,
            service: 'ServiceReference' = None,
            url: str = None,
    ):
        """Create WebhookClientConfig instance."""
        super(WebhookClientConfig, self).__init__(
            api_version='admissionregistration/v1beta1',
            kind='WebhookClientConfig'
        )
        self._properties = {
            'caBundle': ca_bundle if ca_bundle is not None else '',
            'service': service if service is not None else ServiceReference(),
            'url': url if url is not None else '',

        }
        self._types = {
            'caBundle': (str, None),
            'service': (ServiceReference, None),
            'url': (str, None),

        }

    @property
    def ca_bundle(self) -> str:
        """
        `caBundle` is a PEM encoded CA bundle which will be used to
        validate the webhook's server certificate. If unspecified,
        system trust roots on the apiserver are used.
        """
        return self._properties.get('caBundle')

    @ca_bundle.setter
    def ca_bundle(self, value: str):
        """
        `caBundle` is a PEM encoded CA bundle which will be used to
        validate the webhook's server certificate. If unspecified,
        system trust roots on the apiserver are used.
        """
        self._properties['caBundle'] = value

    @property
    def service(self) -> 'ServiceReference':
        """
        `service` is a reference to the service for this webhook.
        Either `service` or `url` must be specified.

        If the webhook is running within the cluster, then you
        should use `service`.

        Port 443 will be used if it is open, otherwise it is an
        error.
        """
        return self._properties.get('service')

    @service.setter
    def service(self, value: typing.Union['ServiceReference', dict]):
        """
        `service` is a reference to the service for this webhook.
        Either `service` or `url` must be specified.

        If the webhook is running within the cluster, then you
        should use `service`.

        Port 443 will be used if it is open, otherwise it is an
        error.
        """
        if isinstance(value, dict):
            value = ServiceReference().from_dict(value)
        self._properties['service'] = value

    @property
    def url(self) -> str:
        """
        `url` gives the location of the webhook, in standard URL
        form (`scheme://host:port/path`). Exactly one of `url` or
        `service` must be specified.

        The `host` should not refer to a service running in the
        cluster; use the `service` field instead. The host might be
        resolved via external DNS in some apiservers (e.g., `kube-
        apiserver` cannot resolve in-cluster DNS as that would be a
        layering violation). `host` may also be an IP address.

        Please note that using `localhost` or `127.0.0.1` as a
        `host` is risky unless you take great care to run this
        webhook on all hosts which run an apiserver which might need
        to make calls to this webhook. Such installs are likely to
        be non-portable, i.e., not easy to turn up in a new cluster.

        The scheme must be "https"; the URL must begin with
        "https://".

        A path is optional, and if present may be any string
        permissible in a URL. You may use the path to pass an
        arbitrary string to the webhook, for example, a cluster
        identifier.

        Attempting to use a user or basic auth e.g. "user:password@"
        is not allowed. Fragments ("#...") and query parameters
        ("?...") are not allowed, either.
        """
        return self._properties.get('url')

    @url.setter
    def url(self, value: str):
        """
        `url` gives the location of the webhook, in standard URL
        form (`scheme://host:port/path`). Exactly one of `url` or
        `service` must be specified.

        The `host` should not refer to a service running in the
        cluster; use the `service` field instead. The host might be
        resolved via external DNS in some apiservers (e.g., `kube-
        apiserver` cannot resolve in-cluster DNS as that would be a
        layering violation). `host` may also be an IP address.

        Please note that using `localhost` or `127.0.0.1` as a
        `host` is risky unless you take great care to run this
        webhook on all hosts which run an apiserver which might need
        to make calls to this webhook. Such installs are likely to
        be non-portable, i.e., not easy to turn up in a new cluster.

        The scheme must be "https"; the URL must begin with
        "https://".

        A path is optional, and if present may be any string
        permissible in a URL. You may use the path to pass an
        arbitrary string to the webhook, for example, a cluster
        identifier.

        Attempting to use a user or basic auth e.g. "user:password@"
        is not allowed. Fragments ("#...") and query parameters
        ("?...") are not allowed, either.
        """
        self._properties['url'] = value

    def __enter__(self) -> 'WebhookClientConfig':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



import typing

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_12.meta_v1 import ListMeta
from kuber.v1_12.meta_v1 import ObjectMeta


class PriorityClass(_kuber_definitions.Resource):
    """
    PriorityClass defines mapping from a priority class name to
    the priority integer value. The value can be any valid
    integer.
    """

    def __init__(
            self,
            description: str = None,
            global_default: bool = None,
            metadata: 'ObjectMeta' = None,
            value: int = None,
    ):
        """Create PriorityClass instance."""
        super(PriorityClass, self).__init__(
            api_version='scheduling/v1beta1',
            kind='PriorityClass'
        )
        self._properties = {
            'description': description if description is not None else '',
            'globalDefault': global_default if global_default is not None else None,
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'value': value if value is not None else None,

        }
        self._types = {
            'apiVersion': (str, None),
            'description': (str, None),
            'globalDefault': (bool, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'value': (int, None),

        }

    @property
    def description(self) -> str:
        """
        description is an arbitrary string that usually provides
        guidelines on when this priority class should be used.
        """
        return self._properties.get('description')

    @description.setter
    def description(self, value: str):
        """
        description is an arbitrary string that usually provides
        guidelines on when this priority class should be used.
        """
        self._properties['description'] = value

    @property
    def global_default(self) -> bool:
        """
        globalDefault specifies whether this PriorityClass should be
        considered as the default priority for pods that do not have
        any priority class. Only one PriorityClass can be marked as
        `globalDefault`. However, if more than one PriorityClasses
        exists with their `globalDefault` field set to true, the
        smallest value of such global default PriorityClasses will
        be used as the default priority.
        """
        return self._properties.get('globalDefault')

    @global_default.setter
    def global_default(self, value: bool):
        """
        globalDefault specifies whether this PriorityClass should be
        considered as the default priority for pods that do not have
        any priority class. Only one PriorityClass can be marked as
        `globalDefault`. However, if more than one PriorityClasses
        exists with their `globalDefault` field set to true, the
        smallest value of such global default PriorityClasses will
        be used as the default priority.
        """
        self._properties['globalDefault'] = value

    @property
    def metadata(self) -> 'ObjectMeta':
        """
        Standard object's metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """
        Standard object's metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def value(self) -> int:
        """
        The value of this priority class. This is the actual
        priority that pods receive when they have the name of this
        class in their pod spec.
        """
        return self._properties.get('value')

    @value.setter
    def value(self, value: int):
        """
        The value of this priority class. This is the actual
        priority that pods receive when they have the name of this
        class in their pod spec.
        """
        self._properties['value'] = value

    def create_resource(self, namespace: 'str' = None):
        """
        Creates the PriorityClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'create_namespaced_priority_class',
            'create_priority_class'
        ]

        _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )

    def replace_resource(self, namespace: 'str' = None):
        """
        Replaces the PriorityClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'replace_namespaced_priority_class',
            'replace_priority_class'
        ]

        _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def patch_resource(self, namespace: 'str' = None):
        """
        Patches the PriorityClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'patch_namespaced_priority_class',
            'patch_priority_class'
        ]

        _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def get_resource_status(self, namespace: 'str' = None):
        """This resource does not have a status."""
        pass

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the PriorityClass from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_priority_class',
            'read_priority_class'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the PriorityClass from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_priority_class',
            'delete_priority_class'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.SchedulingV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.SchedulingV1beta1Api(**kwargs)

    def __enter__(self) -> 'PriorityClass':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class PriorityClassList(_kuber_definitions.Collection):
    """
    PriorityClassList is a collection of priority classes.
    """

    def __init__(
            self,
            items: typing.List['PriorityClass'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create PriorityClassList instance."""
        super(PriorityClassList, self).__init__(
            api_version='scheduling/v1beta1',
            kind='PriorityClassList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, PriorityClass),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['PriorityClass']:
        """
        items is the list of PriorityClasses
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['PriorityClass'], typing.List[dict]]
    ):
        """
        items is the list of PriorityClasses
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = PriorityClass().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.SchedulingV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.SchedulingV1beta1Api(**kwargs)

    def __enter__(self) -> 'PriorityClassList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



import importlib
import os

from kuber import versioning

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))


def test_importable():
    """Should be able to import every generated file."""
    root_path = os.path.dirname(MY_DIRECTORY)

    for version in versioning.get_all_versions():
        v = version.label.replace('.', '_')
        path = os.path.join(root_path, v)
        module_names = [m[:-3] for m in os.listdir(path) if m.endswith('.py')]
        for name in module_names:
            m = importlib.import_module('.'.join(['kuber', v, name]))
            assert m is not None, f'Expected kuber.{v}.{m} to be importable.'



import base64
import json
import os
import typing
from collections import defaultdict

import requests
import semver

VERSION_URL = 'https://api.github.com/repos/kubernetes/kubernetes/tags'
DOWNLOAD_URL = (
    'https://raw.githubusercontent.com/kubernetes/kubernetes/{commit}'
    '/api/openapi-spec/swagger.json'
)


class Release(typing.NamedTuple):
    """Data structure for released versions."""

    name: str
    version: semver.VersionInfo
    api_version: str
    commit_sha: str
    data: dict


class ApiGroup(typing.NamedTuple):
    """Data structure for api release versions."""

    name: str
    version: semver.VersionInfo
    latest: Release
    pre: Release


def to_release_version(release_data: dict) -> typing.Optional[Release]:
    """..."""
    name = release_data['name']
    try:
        version = semver.parse_version_info(name[1:])
    except ValueError:
        return None

    return Release(
        name=name,
        version=version,
        api_version=f'{version.major}.{version.minor}',
        commit_sha=release_data['commit']['sha'],
        data=release_data
    )


def get_release_groups() -> typing.List[ApiGroup]:
    """..."""
    entries = []
    headers = {'Authorization': _get_authentication_header()}
    url = VERSION_URL
    while url is not None:
        response = requests.get(url, headers=headers)
        entries += response.json()
        url = response.links.get('next', {}).get('url')

    releases = [to_release_version(entry) for entry in entries]
    grouped = defaultdict(list)
    for r in releases:
        if r is not None:
            grouped[r.api_version].append(r)

    combined = []
    for version, entries in grouped.items():
        ordered = sorted(entries, reverse=True, key=lambda x: x.version)
        combined.append(ApiGroup(
            name=version,
            version=semver.parse_version_info(f'{version}.0'),
            pre=ordered[0],
            latest=next((e for e in ordered if not e.version.prerelease), None)
        ))

    return sorted(combined, reverse=True, key=lambda x: x.version)


def get_latest_versions() -> typing.Dict[str, Release]:
    """..."""
    groups = get_release_groups()
    limit = 4 if groups[0].latest else 5
    releases = {}

    for group in groups[:limit]:
        release = group.latest or group.pre
        label = 'v{}'.format(release.api_version)
        releases[label] = release

    pre = next((g.pre for g in groups if g.pre), None)
    releases['pre'] = pre

    latest = next((g.latest for g in groups if g.latest), None)
    releases['latest'] = latest

    return releases


def update_specs():
    """..."""
    groups = get_release_groups()
    limit = 4 if groups[0].latest else 5
    for group in groups[:limit]:
        release = group.latest or group.pre
        path = download_version(release, f'v{group.name}')
        print(f'Updated v{group.name}: {path}')

    pre = next((g.pre for g in groups if g.pre), None)
    path = download_version(pre, 'pre')
    print(f'Updated pre ({pre.name}): {path}')

    latest = next((g.latest for g in groups if g.latest), None)
    path = download_version(latest, 'latest')
    print(f'Updated latest ({latest.name}): {path}')


def download_version(release: Release, name: str = None) -> str:
    """
    Downloads the pre or latest version of the specified ApiGroup
    to the specs directory, overwriting anything that might already be
    stored there.
    """
    url = DOWNLOAD_URL.format(commit=release.commit_sha)
    response = requests.get(url)
    data = response.json()
    data['kuber'] = {
        'name': release.name,
        'api_version': release.api_version,
        'commit_sha': release.commit_sha,
        'version': str(release.version)
    }

    path = os.path.realpath(os.path.join(
        os.path.dirname(__file__),
        '..', 'specs', f'{name}.json'
    ))
    with open(path, 'wb') as f:
        f.write(json.dumps(data, indent=2).encode())

    return path


def _get_authentication_header():
    """
    Retrieves Github authentication data from the local directory. Expects
    it to be a JSON file containing name and password keys.
    """
    path = os.path.realpath(os.path.join(
        os.path.dirname(__file__),
        '..', 'github-api-auth.json'
    ))
    with open(path) as f:
        data = json.load(f)

    credentials = f'{data["user"]}:{data["password"]}'
    return 'Basic {}'.format(base64.b64encode(credentials.encode()).decode())


def get_local_versions():
    """..."""
    directory = os.path.realpath(os.path.join(
        os.path.dirname(__file__),
        '..', 'specs'
    ))
    results = {}
    for filename in os.listdir(directory):
        if not filename.endswith('.json'):
            continue

        path = os.path.join(directory, filename)
        label = filename[:-5]
        with open(path) as f:
            results[label] = json.load(f)['kuber']

    return results


def check_for_updates() -> typing.Dict[str, Release]:
    """
    Determines if any updates are needed compared to the locally
    stored versions.
    """
    local_versions = get_local_versions()
    updates = {}
    for label, release in get_latest_versions().items():
        local = local_versions.get(label) or {'version': '0.0.0'}
        version = semver.parse_version_info(local['version'])
        if version != release.version:
            updates[label] = (local, release)

    if not updates:
        print('\nNo updates found. Local versions are up-to-date.\n\n')
        return {}

    print('\n=== Release Updates ===')
    for label, update in updates.items():
        current = update[0]['version']
        latest = update[1].version
        print(f'  - {label}: {current} -> {latest}')

    print(f'\n{len(updates)} updates found.\n\n')

    return {k: v[1] for k, v in updates.items()}



import typing

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_11.meta_v1 import ListMeta
from kuber.v1_11.meta_v1 import ObjectMeta
from kuber.v1_11.core_v1 import TopologySelectorTerm


class StorageClass(_kuber_definitions.Resource):
    """
    StorageClass describes the parameters for a class of storage
    for which PersistentVolumes can be dynamically provisioned.

    StorageClasses are non-namespaced; the name of the storage
    class according to etcd is in ObjectMeta.Name.
    """

    def __init__(
            self,
            allow_volume_expansion: bool = None,
            allowed_topologies: typing.List['TopologySelectorTerm'] = None,
            metadata: 'ObjectMeta' = None,
            mount_options: typing.List[str] = None,
            parameters: dict = None,
            provisioner: str = None,
            reclaim_policy: str = None,
            volume_binding_mode: str = None,
    ):
        """Create StorageClass instance."""
        super(StorageClass, self).__init__(
            api_version='storage/v1',
            kind='StorageClass'
        )
        self._properties = {
            'allowVolumeExpansion': allow_volume_expansion if allow_volume_expansion is not None else None,
            'allowedTopologies': allowed_topologies if allowed_topologies is not None else [],
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'mountOptions': mount_options if mount_options is not None else [],
            'parameters': parameters if parameters is not None else {},
            'provisioner': provisioner if provisioner is not None else '',
            'reclaimPolicy': reclaim_policy if reclaim_policy is not None else '',
            'volumeBindingMode': volume_binding_mode if volume_binding_mode is not None else '',

        }
        self._types = {
            'allowVolumeExpansion': (bool, None),
            'allowedTopologies': (list, TopologySelectorTerm),
            'apiVersion': (str, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'mountOptions': (list, str),
            'parameters': (dict, None),
            'provisioner': (str, None),
            'reclaimPolicy': (str, None),
            'volumeBindingMode': (str, None),

        }

    @property
    def allow_volume_expansion(self) -> bool:
        """
        AllowVolumeExpansion shows whether the storage class allow
        volume expand
        """
        return self._properties.get('allowVolumeExpansion')

    @allow_volume_expansion.setter
    def allow_volume_expansion(self, value: bool):
        """
        AllowVolumeExpansion shows whether the storage class allow
        volume expand
        """
        self._properties['allowVolumeExpansion'] = value

    @property
    def allowed_topologies(self) -> typing.List['TopologySelectorTerm']:
        """
        Restrict the node topologies where volumes can be
        dynamically provisioned. Each volume plugin defines its own
        supported topology specifications. An empty
        TopologySelectorTerm list means there is no topology
        restriction. This field is alpha-level and is only honored
        by servers that enable the DynamicProvisioningScheduling
        feature.
        """
        return self._properties.get('allowedTopologies')

    @allowed_topologies.setter
    def allowed_topologies(
            self,
            value: typing.Union[typing.List['TopologySelectorTerm'], typing.List[dict]]
    ):
        """
        Restrict the node topologies where volumes can be
        dynamically provisioned. Each volume plugin defines its own
        supported topology specifications. An empty
        TopologySelectorTerm list means there is no topology
        restriction. This field is alpha-level and is only honored
        by servers that enable the DynamicProvisioningScheduling
        feature.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = TopologySelectorTerm().from_dict(item)
            cleaned.append(item)
        self._properties['allowedTopologies'] = cleaned

    @property
    def metadata(self) -> 'ObjectMeta':
        """
        Standard object's metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """
        Standard object's metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def mount_options(self) -> typing.List[str]:
        """
        Dynamically provisioned PersistentVolumes of this storage
        class are created with these mountOptions, e.g. ["ro",
        "soft"]. Not validated - mount of the PVs will simply fail
        if one is invalid.
        """
        return self._properties.get('mountOptions')

    @mount_options.setter
    def mount_options(self, value: typing.List[str]):
        """
        Dynamically provisioned PersistentVolumes of this storage
        class are created with these mountOptions, e.g. ["ro",
        "soft"]. Not validated - mount of the PVs will simply fail
        if one is invalid.
        """
        self._properties['mountOptions'] = value

    @property
    def parameters(self) -> dict:
        """
        Parameters holds the parameters for the provisioner that
        should create volumes of this storage class.
        """
        return self._properties.get('parameters')

    @parameters.setter
    def parameters(self, value: dict):
        """
        Parameters holds the parameters for the provisioner that
        should create volumes of this storage class.
        """
        self._properties['parameters'] = value

    @property
    def provisioner(self) -> str:
        """
        Provisioner indicates the type of the provisioner.
        """
        return self._properties.get('provisioner')

    @provisioner.setter
    def provisioner(self, value: str):
        """
        Provisioner indicates the type of the provisioner.
        """
        self._properties['provisioner'] = value

    @property
    def reclaim_policy(self) -> str:
        """
        Dynamically provisioned PersistentVolumes of this storage
        class are created with this reclaimPolicy. Defaults to
        Delete.
        """
        return self._properties.get('reclaimPolicy')

    @reclaim_policy.setter
    def reclaim_policy(self, value: str):
        """
        Dynamically provisioned PersistentVolumes of this storage
        class are created with this reclaimPolicy. Defaults to
        Delete.
        """
        self._properties['reclaimPolicy'] = value

    @property
    def volume_binding_mode(self) -> str:
        """
        VolumeBindingMode indicates how PersistentVolumeClaims
        should be provisioned and bound.  When unset,
        VolumeBindingImmediate is used. This field is alpha-level
        and is only honored by servers that enable the
        VolumeScheduling feature.
        """
        return self._properties.get('volumeBindingMode')

    @volume_binding_mode.setter
    def volume_binding_mode(self, value: str):
        """
        VolumeBindingMode indicates how PersistentVolumeClaims
        should be provisioned and bound.  When unset,
        VolumeBindingImmediate is used. This field is alpha-level
        and is only honored by servers that enable the
        VolumeScheduling feature.
        """
        self._properties['volumeBindingMode'] = value

    def create_resource(self, namespace: 'str' = None):
        """
        Creates the StorageClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'create_namespaced_storage_class',
            'create_storage_class'
        ]

        _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )

    def replace_resource(self, namespace: 'str' = None):
        """
        Replaces the StorageClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'replace_namespaced_storage_class',
            'replace_storage_class'
        ]

        _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def patch_resource(self, namespace: 'str' = None):
        """
        Patches the StorageClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'patch_namespaced_storage_class',
            'patch_storage_class'
        ]

        _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def get_resource_status(self, namespace: 'str' = None):
        """This resource does not have a status."""
        pass

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the StorageClass from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_storage_class',
            'read_storage_class'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the StorageClass from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_storage_class',
            'delete_storage_class'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.StorageV1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.StorageV1Api(**kwargs)

    def __enter__(self) -> 'StorageClass':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class StorageClassList(_kuber_definitions.Collection):
    """
    StorageClassList is a collection of storage classes.
    """

    def __init__(
            self,
            items: typing.List['StorageClass'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create StorageClassList instance."""
        super(StorageClassList, self).__init__(
            api_version='storage/v1',
            kind='StorageClassList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, StorageClass),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['StorageClass']:
        """
        Items is the list of StorageClasses
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['StorageClass'], typing.List[dict]]
    ):
        """
        Items is the list of StorageClasses
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = StorageClass().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.StorageV1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.StorageV1Api(**kwargs)

    def __enter__(self) -> 'StorageClassList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



import os

import kuber
from kuber.latest import apps_v1


def test_zero_scale_deployment():
    """Should preserve a 0 set on the scale for a deployment."""
    d = apps_v1.Deployment()
    d.spec.replicas = 0
    result = d.to_yaml()
    assert 'replicas: 0' in result


def test_zero_replicas_from_file():
    """Should preserve a 0 replicas on a deployment loaded from yaml."""
    path = os.path.realpath(os.path.join(
        os.path.dirname(__file__),
        'deployment.yaml'
    ))
    bundle = kuber.from_file(path)
    assert 'replicas: 0' in bundle.render_yaml_bundle()



import argparse
import os
import typing

import kuber_maker
from kuber_maker import documenting
from kuber_maker import initializer
from kuber_maker import packaging
from kuber_maker import parsing
from kuber_maker import render
from kuber_maker import updater


def generate(version: str):
    """..."""
    spec = kuber_maker.load_spec(version)
    initializer.empty(version)
    initializer.create_subpackages(version, spec.definitions)
    entities = parsing.parse_definitions(version, spec.definitions)

    for package in entities.packages.keys():
        contents = render.render_package(package, entities)
        path = kuber_maker.directory_of_package(package)

        with open(f'{path}.py', 'wb') as f:
            f.write(contents.encode())
        print(f'[CREATED]: {path}.py')

    contents = render.render_root_package(version, spec)
    path = kuber_maker.get_path(version, '__init__.py')
    with open(path, 'w') as f:
        f.write(contents)
    print(f'[RENDERED]: {path}')


def generate_all():
    """..."""
    directory = os.path.realpath(os.path.join(
        os.path.dirname(__file__),
        '..', 'specs'
    ))
    spec_files = [f for f in os.listdir(directory) if f.endswith('.json')]
    for filename in spec_files:
        version = filename[:-5]
        print(f'\n\n=== {version} ===')
        generate(version)


def main(args: typing.List[str] = None):
    """..."""
    parser = argparse.ArgumentParser(prog='kuber_maker')
    subs = parser.add_subparsers(dest='action')

    p = subs.add_parser('build')
    p.add_argument('version', nargs='?')
    p.add_argument('--all', action='store_true')

    subs.add_parser('update')
    subs.add_parser('status')
    subs.add_parser('docs')
    subs.add_parser('apidocs')
    subs.add_parser('release')

    arguments = parser.parse_args(args=args)
    if arguments.action == 'status':
        return updater.check_for_updates()
    if arguments.action == 'update':
        return updater.update_specs()
    if arguments.action == 'release':
        return packaging.deploy_release()
    if arguments.action == 'docs':
        return documenting.build_docs()
    if arguments.action == 'apidocs':
        return documenting.update_api_docs()
    if arguments.version:
        return generate(arguments.version)
    elif arguments.all:
        return generate_all()

    raise ValueError('Specify a version or select all')



import os

import kuber
from kuber.latest import apiextensions_v1beta1

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))


def test_custom_resource_definition():
    """
    Should not cause a recursion error while loading a CRD yaml definition.
    For more details on the source of this scenario see:
    https://github.com/sernst/kuber/issues/1
    """
    bundle = (
        kuber.create_bundle(kubernetes_version='latest')
        .add_file(os.path.join(MY_DIRECTORY, 'crd.yaml'))
    )
    assert len(bundle.resources) == 1

    crd: apiextensions_v1beta1.CustomResourceDefinition = bundle.resources[0]
    assert crd.spec.validation.open_apiv3_schema.not_ is None, """
        Expect the self-referential property value to be None by default.
        """



import os
import subprocess
import shutil

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))
ROOT_DIRECTORY = os.path.realpath(os.path.join(MY_DIRECTORY, '..'))


def update_api_docs():
    """Rebuilds the API documentation for the kuber library."""
    directory = os.path.join(ROOT_DIRECTORY, 'docs', 'modules')
    if os.path.exists(directory):
        shutil.rmtree(directory)

    os.chdir(ROOT_DIRECTORY)
    cmd = ['sphinx-apidoc', 'kuber', '-o', directory]
    subprocess.run(cmd).check_returncode()

    build_docs()


def build_docs():
    """Generates local sphinx docs."""
    path = os.path.join(ROOT_DIRECTORY, '_build')
    cmd = ['sphinx-build', '-M', 'html', ROOT_DIRECTORY, path]
    subprocess.run(cmd).check_returncode()



from pytest import mark

from kuber import definitions
from kuber.latest import core_v1


def test_definition_serialization():
    """Should create a Definition object and serialize as expected."""
    d = definitions.Definition('1.11', 'Deployment')
    d._properties = {'foo': 1, 'bar': None, 'baz': 'hello'}
    d._types = {'bar': (float, None), 'baz': (str, None)}

    serialized = d.to_dict()
    assert {'baz': 'hello'} == serialized
    assert d.kuber_uid is not None


DESERIALIZATION_SCENARIOS = [
    {'fooBar': 1, 'barFoo': None, 'barBaz': 'hello', 'BAZQux': True},
    {'foo_bar': 1, 'bar_foo': None, 'bar_baz': 'hello', 'baz_qux': True},
    {'FOOBAR': 1, 'bar_foo': None, 'BARBAZ': 'hello', 'BAZQUX': True},
]


@mark.parametrize('scenario', DESERIALIZATION_SCENARIOS)
def test_definition_deserialization(scenario: dict):
    """Should create a Definition object populated from the serialized data."""
    d = definitions.Definition('1.11', 'Deployment')
    d._types = {
        'fooBar': (float, None),
        'barBaz': (str, None),
        'BAZQux': (bool, None)
    }
    expected = {'barBaz': 'hello', 'fooBar': 1, 'BAZQux': True}
    assert expected == d.from_dict({'a': 1, 'b': 2, **scenario})._properties


def test_collection_yaml_serialization():
    """Should serialize the Collection to a YAML string."""
    pod_list = core_v1.PodList()
    pod_list.metadata.resource_version = 'foo'
    assert 'foo' in pod_list.to_yaml()


def test_collection_json_serialization():
    """Should serialize the Collection to a JSON string."""
    namespace_list = core_v1.NamespaceList()
    namespace_list.metadata.resource_version = 'foo'
    assert 'foo' in namespace_list.to_json()


SCENARIOS = [
    ('foo_bar_baz', 'fooBarBaz'),
    ('foo-bar-baz', 'fooBarBaz'),
    ('fooBarBaz', 'fooBarBaz')
]


@mark.parametrize('source, expected', SCENARIOS)
def test_to_camel_case(source: str, expected: str):
    """Should convert the source string to the expected camelCase result."""
    assert expected == definitions.to_camel_case(source)



import json
import os
import typing
import semver


class ApiSpec(typing.NamedTuple):
    """Data structure for high-level API info."""

    # Title from the swagger info object spec.
    title: str
    # Version as specified by the GitHub release name.
    version: str
    # Version from the info object in the Swagger spec. This make not
    # always be accurate in pre-releases as it appears to linger on the
    # last stable version.
    info_version: str
    # Semantic version as specified by the GitHub release name.
    version_info: semver.VersionInfo
    # Paths specified in the swagger definition.
    api_paths: typing.Dict[str, dict]
    # Definitions specified in the swagger definition.
    definitions: typing.Dict[str, dict]
    # The sha commit identifier that contains the swagger definition
    # used to generate this spec
    commit_sha: str


class Import(typing.NamedTuple):
    """..."""

    target: str
    package: str
    api_path: str
    reference: str


class DataType(typing.NamedTuple):
    """..."""

    api_type: str
    python_type: str
    constructor: str
    type_hint: str
    code_import: typing.Optional[Import] = None
    item_type: typing.Optional[str] = None
    # A special case type-hint specifically for use when specifying a
    # setter function argument type.
    setter_type_hint: str = None


class Property(typing.NamedTuple):
    """..."""

    name: str
    data_type: DataType
    description: str
    source: dict


class Entity(typing.NamedTuple):
    """..."""

    api_path: str
    api_version: str
    package: str
    code_path: str
    class_name: str
    description: str
    properties: typing.Dict[str, Property]
    data_type: str
    source: dict
    imports: typing.List[Import]
    is_resource: bool = False


class EntityPackage(typing.NamedTuple):
    """..."""

    package: str
    entities: typing.Dict[str, Entity]


class AllEntities(typing.NamedTuple):
    """..."""

    version: str
    packages: typing.Dict[str, EntityPackage]


def get_package(version, *args) -> str:
    """..."""
    v = version.replace('.', '_')
    return '.'.join(['kuber', f'{v}', *args])


def directory_of_package(package: str) -> str:
    """..."""
    return os.path.realpath(os.path.join(
        os.path.dirname(__file__),
        '..',
        *package.split('.')
    ))


def to_kuber_package(version: str, kubernetes_api_path: str) -> str:
    """..."""
    return get_package(version, *_to_kuber_hierarchy(kubernetes_api_path))


def get_path(version, *args) -> str:
    """..."""
    v = version.replace('.', '_')
    return os.path.realpath(os.path.join(
        os.path.dirname(__file__),
        '..', 'kuber', f'{v}', *args
    ))


def to_kuber_path(version: str, kubernetes_api_path: str) -> str:
    """..."""
    return get_path(version, *_to_kuber_hierarchy(kubernetes_api_path))


def import_from_reference(version: str, reference_path: str) -> Import:
    """..."""
    path = reference_path.replace('#/definitions/', '')
    package = to_kuber_package(version, path)
    return Import(
        target=reference_path.rsplit('.', 1)[-1],
        package=package,
        api_path=path,
        reference=reference_path
    )


def load_spec(version: str) -> ApiSpec:
    """..."""
    path = os.path.realpath(os.path.join(
        os.path.dirname(__file__),
        '..', 'specs', f'{version}.json'
    ))
    with open(path, 'rb') as f:
        contents = f.read()
    raw = json.loads(contents)
    return ApiSpec(
        title=raw['info']['title'],
        info_version=raw['info']['version'],
        version=raw['kuber']['name'],
        version_info=semver.parse_version_info(raw['kuber']['version']),
        api_paths=raw['paths'],
        definitions=raw['definitions'],
        commit_sha=raw['kuber']['commit_sha']
    )


def _to_kuber_hierarchy(kubernetes_api_path: str) -> typing.List[str]:
    """..."""
    path = (
        kubernetes_api_path
        .replace('-', '_')
        .replace('io.k8s.api.', '')
        .replace('io.k8s.', '')
        .replace('kube_aggregator.pkg.apis.', '')
        .replace('apimachinery.pkg.apis.', '')
        .replace('apiextensions_apiserver.pkg.apis.', '')
        .replace('pkg.apis.', '')
        .replace('pkg.', '')
    )

    # Remove file extension and split into hierarchy list.
    parts = path.rsplit('.')[:-1]

    # Combine the api versions, `apps, v1` into `apps_v1`.
    combined = '_'.join(parts[-2:])

    return parts[:-2] + [combined]



from kuber.latest import batch_v1beta1


def test_cron_suspend():
    """Should preserve suspend value when false."""
    cron = batch_v1beta1.CronJob(
        spec=batch_v1beta1.CronJobSpec(
            suspend=True
        )
    )
    before = cron.to_yaml()
    cron.spec.suspend = False
    after = cron.to_yaml()
    assert 'suspend: true' in before
    assert 'suspend: false' in after



import shutil
import os
import kuber_maker


def empty(version: str):
    """..."""
    path = kuber_maker.get_path(version)
    if os.path.exists(path):
        shutil.rmtree(path)


def create_subpackages(version: str, definitions: dict):
    paths = []
    for api_path in definitions.keys():
        module_path = kuber_maker.to_kuber_path(version, api_path)
        path = os.path.dirname(module_path)

        if os.path.basename(path) in ('kubernetes', 'apimachinery'):
            # These sub-packages are empty because the resources
            # here have been moved and the only items remaining are
            # the type definitions that will be skipped Time, Quantity,
            # and IntOrString.
            continue

        paths.append(path)

        if not os.path.exists(path):
            os.makedirs(path)
        _populate_package_inits(path)

    return paths


def _populate_package_inits(path: str):
    """..."""
    init_path = os.path.join(path, '__init__.py')
    while not os.path.exists(init_path):
        with open(init_path, 'w') as f:
            f.write('')
        init_path = os.path.join(
            os.path.dirname(os.path.dirname(init_path)),
            '__init__.py'
        )



import typing as _typing
import yaml as _yaml

from kuber import versioning as _versioning
from kuber.interface import CommandAction  # noqa
from kuber.definitions import Collection  # noqa
from kuber.definitions import Definition  # noqa
from kuber.definitions import Resource  # noqa
from kuber.kube_api import get_version_from_cluster  # noqa
from kuber.kube_api import load_access_config  # noqa
from kuber.management import ResourceBundle
from kuber.management.creation import from_dict  # noqa
from kuber.management.creation import from_json_file  # noqa
from kuber.management.creation import from_yaml  # noqa
from kuber.management.creation import from_yaml_file  # noqa
from kuber.management.creation import from_yaml_file_multiple  # noqa
from kuber.management.creation import from_yaml_multiple  # noqa
from kuber.management.creation import new_resource  # noqa
from kuber.versioning import KubernetesVersion  # noqa

#: kuber library version.
__version__ = '1.11.2'

#: The loader used when loading yaml via pyyaml. This can be overridden
#: in cases where a different Loader is preferred.
yaml_loader = _yaml.Loader

#: All currently supported versions that exist within this installation
#: of the kuber library.
available_versions: _typing.List[KubernetesVersion] = (
    _versioning.get_all_versions()
)

#: The most recent kubernetes version available within the library, which
#: can be used to avoid hard-coded versions when creating resource bundles.
latest_kube_version: KubernetesVersion = (
    _versioning.get_latest_version(stable=True)
)

#: Type that accepts either a `KubernetesVersion` object or a string
#: representation of that version, e.g. 'v1.14' or 'latest'.
VersionLabel = _typing.Union[KubernetesVersion, str]


def create_bundle(
        kubernetes_version: VersionLabel = None,
        bundle_name: str = None
) -> ResourceBundle:
    """
    Creates a `ResourceBundle` instance initialized to define resources from
    the given Kubernetes version.

    :param kubernetes_version:
        Kubernetes version in the form MAJOR.MINOR (no patch) that should be
        used by the bundle when creating and operating on Kubernetes Resource
        objects (e.g. `Deployment`, `Pod`, etc.). If not specified the most
        recent version will be used.
    :param bundle_name:
        A name to associated with the resource bundle. If not specified, a
        randomized name will be generated instead.
    :return:
        Initialized `ResourceBundle`.
    """
    return ResourceBundle(bundle_name, kubernetes_version)


def from_directory_files(
        directory: str,
        filenames: _typing.Iterable = None,
        kubernetes_version: VersionLabel = None,
        bundle_name: str = None
) -> ResourceBundle:
    """
    Creates a `ResourceBundle` object from all of the resource
    configuration files specified by the ``filenames`` argument from
    within the specified directory.

    :param directory:
        Path of the directory where the Kubernetes Resource configuration files
        to be loaded are stored.
    :param filenames:
        A list of filenames in the given directory and the order in
        which they should be added.
    :param kubernetes_version:
        Kubernetes version in the form MAJOR.MINOR (no patch) that should be
        used by the bundle when creating and operating on Kubernetes Resource
        objects (e.g. `Deployment`, `Pod`, etc.). If not specified the most
        recent version will be used.
    :param bundle_name:
        A name to associated with the resource bundle. If not specified, a
        randomized name will be generated instead.
    :return:
        Initialized `ResourceBundle` populated with all of the Kubernetes
        Resource objects loaded from the specified directory.
    """
    bundle = ResourceBundle(bundle_name, kubernetes_version)
    return bundle.add_directory_files(directory, filenames=filenames)


def from_directory(
        directory: str,
        recursive: bool = False,
        kubernetes_version: VersionLabel = None,
        bundle_name: str = None
) -> ResourceBundle:
    """
    Creates a `ResourceBundle` object from all of the yaml and json
    configuration files within the specified directory.

    :param directory:
        Path of the directory where the Kubernetes Resource configuration files
        to be loaded are stored.
    :param recursive:
        Whether or not to recursively load resources files from descendent
        folders within the specified directory.
    :param kubernetes_version:
        Kubernetes version in the form MAJOR.MINOR (no patch) that should be
        used by the bundle when creating and operating on Kubernetes Resource
        objects (e.g. `Deployment`, `Pod`, etc.). If not specified the most
        recent version will be used.
    :param bundle_name:
        A name to associated with the resource bundle. If not specified, a
        randomized name will be generated instead.
    :return:
        Initialized `ResourceBundle` populated with all of the Kubernetes
        Resource objects loaded from the specified directory.
    """
    bundle = ResourceBundle(bundle_name, kubernetes_version)
    return bundle.add_directory(directory, recursive=recursive)


def from_file(
        path: str,
        kubernetes_version: VersionLabel = None,
        bundle_name: str = None
) -> ResourceBundle:
    """
    Creates a `ResourceBundle` object and populates it with the Kubernetes
    Resource configuration specified by the `path` argument.

    :param path:
        Path to the resource configuration file that should be loaded as a
        resource into the newly created resource bundle.
    :param kubernetes_version:
        Kubernetes version in the form MAJOR.MINOR (no patch) that should be
        used by the bundle when creating and operating on Kubernetes Resource
        objects (e.g. `Deployment`, `Pod`, etc.). If not specified the most
        recent version will be used.
    :param bundle_name:
        A name to associated with the resource bundle. If not specified, a
        randomized name will be generated instead.
    :return:
        Initialized `ResourceBundle` populated with the Kubernetes Resource
        specified by the configuration file of the given `path`.
    """
    bundle = ResourceBundle(bundle_name, kubernetes_version)
    return bundle.add_file(path)


def cli(
        callback: _typing.Callable[['CommandAction'], _typing.Any],
        kubernetes_version: VersionLabel = None,
        bundle_name: str = None,
        arguments: _typing.List[str] = None
) -> 'management.ResourceBundle':
    """
    Creates an empty bundle configured with the optionally specified
    Kubernetes version and bundle name and immediately invokes the command
    line interface for the bundle, but invoking the specified callback before
    executing the command. The callback is the opportunity to configure the
    bundle.

    :param callback:
        Callback that will be executed prior to the command line action
        execution. Use this callback to configure the bundle using input
        from the command line (including custom command arguments) before
        the cli command action is invoked.
    :param kubernetes_version:
        Kubernetes version in the form MAJOR.MINOR (no patch) that should be
        used by the bundle when creating and operating on Kubernetes Resource
        objects (e.g. `Deployment`, `Pod`, etc.). If not specified the most
        recent version will be used.
    :param bundle_name:
        A name to associated with the resource bundle. If not specified, a
        randomized name will be generated instead.
    :param arguments:
        Optional command line arguments to parse and use in the callback
        and then in the cli action execution. If omitted the arguments
        will be parsed from the sys.argv values provided on the command line.
    """
    bundle = create_bundle(
        kubernetes_version=kubernetes_version,
        bundle_name=bundle_name
    )
    bundle.cli.invoke(
        callback=callback,
        arguments=arguments
    )
    return bundle



import typing
import datetime as _datetime

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_17.core_v1 import Container
from kuber.v1_17.core_v1 import ContainerPort
from kuber.v1_17.core_v1 import EnvFromSource
from kuber.v1_17.core_v1 import EnvVar
from kuber.v1_17.batch_v1 import JobSpec
from kuber.v1_17.core_v1 import Lifecycle
from kuber.v1_17.meta_v1 import ListMeta
from kuber.v1_17.meta_v1 import ObjectMeta
from kuber.v1_17.core_v1 import ObjectReference
from kuber.v1_17.core_v1 import Probe
from kuber.v1_17.core_v1 import ResourceRequirements
from kuber.v1_17.core_v1 import SecurityContext
from kuber.v1_17.meta_v1 import Status
from kuber.v1_17.meta_v1 import StatusDetails
from kuber.v1_17.core_v1 import VolumeDevice
from kuber.v1_17.core_v1 import VolumeMount


class CronJob(_kuber_definitions.Resource):
    """
    CronJob represents the configuration of a single cron job.
    """

    def __init__(
            self,
            metadata: 'ObjectMeta' = None,
            spec: 'CronJobSpec' = None,
            status: 'CronJobStatus' = None,
    ):
        """Create CronJob instance."""
        super(CronJob, self).__init__(
            api_version='batch/v1beta1',
            kind='CronJob'
        )
        self._properties = {
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'spec': spec if spec is not None else CronJobSpec(),
            'status': status if status is not None else CronJobStatus(),

        }
        self._types = {
            'apiVersion': (str, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'spec': (CronJobSpec, None),
            'status': (CronJobStatus, None),

        }

    @property
    def metadata(self) -> 'ObjectMeta':
        """
        Standard object's metadata. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """
        Standard object's metadata. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def spec(self) -> 'CronJobSpec':
        """
        Specification of the desired behavior of a cron job,
        including the schedule. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#spec-and-status
        """
        return self._properties.get('spec')

    @spec.setter
    def spec(self, value: typing.Union['CronJobSpec', dict]):
        """
        Specification of the desired behavior of a cron job,
        including the schedule. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#spec-and-status
        """
        if isinstance(value, dict):
            value = CronJobSpec().from_dict(value)
        self._properties['spec'] = value

    @property
    def status(self) -> 'CronJobStatus':
        """
        Current status of a cron job. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#spec-and-status
        """
        return self._properties.get('status')

    @status.setter
    def status(self, value: typing.Union['CronJobStatus', dict]):
        """
        Current status of a cron job. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#spec-and-status
        """
        if isinstance(value, dict):
            value = CronJobStatus().from_dict(value)
        self._properties['status'] = value

    def append_container(
        self,
        args: typing.List[str] = _kuber_definitions.UNCHANGED_VALUE,
        command: typing.List[str] = _kuber_definitions.UNCHANGED_VALUE,
        env: typing.List['EnvVar'] = _kuber_definitions.UNCHANGED_VALUE,
        env_from: typing.List['EnvFromSource'] = _kuber_definitions.UNCHANGED_VALUE,
        image: str = _kuber_definitions.UNCHANGED_VALUE,
        image_pull_policy: str = _kuber_definitions.UNCHANGED_VALUE,
        lifecycle: 'Lifecycle' = _kuber_definitions.UNCHANGED_VALUE,
        liveness_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        name: str = _kuber_definitions.UNCHANGED_VALUE,
        ports: typing.List['ContainerPort'] = _kuber_definitions.UNCHANGED_VALUE,
        readiness_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        resources: 'ResourceRequirements' = _kuber_definitions.UNCHANGED_VALUE,
        security_context: 'SecurityContext' = _kuber_definitions.UNCHANGED_VALUE,
        startup_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        stdin: bool = _kuber_definitions.UNCHANGED_VALUE,
        stdin_once: bool = _kuber_definitions.UNCHANGED_VALUE,
        termination_message_path: str = _kuber_definitions.UNCHANGED_VALUE,
        termination_message_policy: str = _kuber_definitions.UNCHANGED_VALUE,
        tty: bool = _kuber_definitions.UNCHANGED_VALUE,
        volume_devices: typing.List['VolumeDevice'] = _kuber_definitions.UNCHANGED_VALUE,
        volume_mounts: typing.List['VolumeMount'] = _kuber_definitions.UNCHANGED_VALUE,
        working_dir: str = _kuber_definitions.UNCHANGED_VALUE,
    ) -> 'CronJob':
        """Adds a container object within the specified resource."""
        values = {
            'args': args,
            'command': command,
            'env': env,
            'env_from': env_from,
            'image': image,
            'image_pull_policy': image_pull_policy,
            'lifecycle': lifecycle,
            'liveness_probe': liveness_probe,
            'name': name,
            'ports': ports,
            'readiness_probe': readiness_probe,
            'resources': resources,
            'security_context': security_context,
            'startup_probe': startup_probe,
            'stdin': stdin,
            'stdin_once': stdin_once,
            'termination_message_path': termination_message_path,
            'termination_message_policy': termination_message_policy,
            'tty': tty,
            'volume_devices': volume_devices,
            'volume_mounts': volume_mounts,
            'working_dir': working_dir,
        }
        self.spec.job_template.spec.template.spec.containers.append(Container(**{
            k: v
            for k, v in values.items()
            if v != _kuber_definitions.UNCHANGED_VALUE
        }))
        return self

    def get_container(self, name: str) -> typing.Optional['Container']:
        """
        Fetch a container definition within this resource by name if such a
        container definition exists. Return None if no container definition
        by that name exists.
        """
        return next((c for c in self.spec.job_template.spec.template.spec.containers if c.name == name), None)

    def get_containers(self) -> typing.List['Container']:
        """
        Returns the list of containers stored in this resource if any such
        containers exist.
        """
        return self.spec.job_template.spec.template.spec.containers

    def create_resource(
            self,
            namespace: 'str' = None
    ) -> 'CronJobStatus':
        """
        Creates the CronJob in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the create is complete.
        """
        names = [
            'create_namespaced_cron_job',
            'create_cron_job'
        ]

        response = _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )
        return (
            CronJobStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def replace_resource(
            self,
            namespace: 'str' = None
    ) -> 'CronJobStatus':
        """
        Replaces the CronJob in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the replace is complete.
        """
        names = [
            'replace_namespaced_cron_job',
            'replace_cron_job'
        ]

        response = _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )
        return (
            CronJobStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def patch_resource(
            self,
            namespace: 'str' = None
    ) -> 'CronJobStatus':
        """
        Patches the CronJob in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the replace is complete.
        """
        names = [
            'patch_namespaced_cron_job',
            'patch_cron_job'
        ]

        response = _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )
        return (
            CronJobStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def get_resource_status(
            self,
            namespace: 'str' = None
    ) -> 'CronJobStatus':
        """
        Returns status information about the given resource within the cluster.
        """
        names = [
            'read_namespaced_cron_job',
            'read_cron_job'
        ]

        response = _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )
        return (
            CronJobStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the CronJob from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_cron_job',
            'read_cron_job'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the CronJob from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_cron_job',
            'delete_cron_job'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.BatchV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.BatchV1beta1Api(**kwargs)

    def __enter__(self) -> 'CronJob':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CronJobList(_kuber_definitions.Collection):
    """
    CronJobList is a collection of cron jobs.
    """

    def __init__(
            self,
            items: typing.List['CronJob'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create CronJobList instance."""
        super(CronJobList, self).__init__(
            api_version='batch/v1beta1',
            kind='CronJobList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, CronJob),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['CronJob']:
        """
        items is the list of CronJobs.
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['CronJob'], typing.List[dict]]
    ):
        """
        items is the list of CronJobs.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = CronJob().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.BatchV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.BatchV1beta1Api(**kwargs)

    def __enter__(self) -> 'CronJobList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CronJobSpec(_kuber_definitions.Definition):
    """
    CronJobSpec describes how the job execution will look like
    and when it will actually run.
    """

    def __init__(
            self,
            concurrency_policy: str = None,
            failed_jobs_history_limit: int = None,
            job_template: 'JobTemplateSpec' = None,
            schedule: str = None,
            starting_deadline_seconds: int = None,
            successful_jobs_history_limit: int = None,
            suspend: bool = None,
    ):
        """Create CronJobSpec instance."""
        super(CronJobSpec, self).__init__(
            api_version='batch/v1beta1',
            kind='CronJobSpec'
        )
        self._properties = {
            'concurrencyPolicy': concurrency_policy if concurrency_policy is not None else '',
            'failedJobsHistoryLimit': failed_jobs_history_limit if failed_jobs_history_limit is not None else None,
            'jobTemplate': job_template if job_template is not None else JobTemplateSpec(),
            'schedule': schedule if schedule is not None else '',
            'startingDeadlineSeconds': starting_deadline_seconds if starting_deadline_seconds is not None else None,
            'successfulJobsHistoryLimit': successful_jobs_history_limit if successful_jobs_history_limit is not None else None,
            'suspend': suspend if suspend is not None else None,

        }
        self._types = {
            'concurrencyPolicy': (str, None),
            'failedJobsHistoryLimit': (int, None),
            'jobTemplate': (JobTemplateSpec, None),
            'schedule': (str, None),
            'startingDeadlineSeconds': (int, None),
            'successfulJobsHistoryLimit': (int, None),
            'suspend': (bool, None),

        }

    @property
    def concurrency_policy(self) -> str:
        """
        Specifies how to treat concurrent executions of a Job. Valid
        values are: - "Allow" (default): allows CronJobs to run
        concurrently; - "Forbid": forbids concurrent runs, skipping
        next run if previous run hasn't finished yet; - "Replace":
        cancels currently running job and replaces it with a new one
        """
        return self._properties.get('concurrencyPolicy')

    @concurrency_policy.setter
    def concurrency_policy(self, value: str):
        """
        Specifies how to treat concurrent executions of a Job. Valid
        values are: - "Allow" (default): allows CronJobs to run
        concurrently; - "Forbid": forbids concurrent runs, skipping
        next run if previous run hasn't finished yet; - "Replace":
        cancels currently running job and replaces it with a new one
        """
        self._properties['concurrencyPolicy'] = value

    @property
    def failed_jobs_history_limit(self) -> int:
        """
        The number of failed finished jobs to retain. This is a
        pointer to distinguish between explicit zero and not
        specified. Defaults to 1.
        """
        return self._properties.get('failedJobsHistoryLimit')

    @failed_jobs_history_limit.setter
    def failed_jobs_history_limit(self, value: int):
        """
        The number of failed finished jobs to retain. This is a
        pointer to distinguish between explicit zero and not
        specified. Defaults to 1.
        """
        self._properties['failedJobsHistoryLimit'] = value

    @property
    def job_template(self) -> 'JobTemplateSpec':
        """
        Specifies the job that will be created when executing a
        CronJob.
        """
        return self._properties.get('jobTemplate')

    @job_template.setter
    def job_template(self, value: typing.Union['JobTemplateSpec', dict]):
        """
        Specifies the job that will be created when executing a
        CronJob.
        """
        if isinstance(value, dict):
            value = JobTemplateSpec().from_dict(value)
        self._properties['jobTemplate'] = value

    @property
    def schedule(self) -> str:
        """
        The schedule in Cron format, see
        https://en.wikipedia.org/wiki/Cron.
        """
        return self._properties.get('schedule')

    @schedule.setter
    def schedule(self, value: str):
        """
        The schedule in Cron format, see
        https://en.wikipedia.org/wiki/Cron.
        """
        self._properties['schedule'] = value

    @property
    def starting_deadline_seconds(self) -> int:
        """
        Optional deadline in seconds for starting the job if it
        misses scheduled time for any reason.  Missed jobs
        executions will be counted as failed ones.
        """
        return self._properties.get('startingDeadlineSeconds')

    @starting_deadline_seconds.setter
    def starting_deadline_seconds(self, value: int):
        """
        Optional deadline in seconds for starting the job if it
        misses scheduled time for any reason.  Missed jobs
        executions will be counted as failed ones.
        """
        self._properties['startingDeadlineSeconds'] = value

    @property
    def successful_jobs_history_limit(self) -> int:
        """
        The number of successful finished jobs to retain. This is a
        pointer to distinguish between explicit zero and not
        specified. Defaults to 3.
        """
        return self._properties.get('successfulJobsHistoryLimit')

    @successful_jobs_history_limit.setter
    def successful_jobs_history_limit(self, value: int):
        """
        The number of successful finished jobs to retain. This is a
        pointer to distinguish between explicit zero and not
        specified. Defaults to 3.
        """
        self._properties['successfulJobsHistoryLimit'] = value

    @property
    def suspend(self) -> bool:
        """
        This flag tells the controller to suspend subsequent
        executions, it does not apply to already started executions.
        Defaults to false.
        """
        return self._properties.get('suspend')

    @suspend.setter
    def suspend(self, value: bool):
        """
        This flag tells the controller to suspend subsequent
        executions, it does not apply to already started executions.
        Defaults to false.
        """
        self._properties['suspend'] = value

    def append_container(
        self,
        args: typing.List[str] = _kuber_definitions.UNCHANGED_VALUE,
        command: typing.List[str] = _kuber_definitions.UNCHANGED_VALUE,
        env: typing.List['EnvVar'] = _kuber_definitions.UNCHANGED_VALUE,
        env_from: typing.List['EnvFromSource'] = _kuber_definitions.UNCHANGED_VALUE,
        image: str = _kuber_definitions.UNCHANGED_VALUE,
        image_pull_policy: str = _kuber_definitions.UNCHANGED_VALUE,
        lifecycle: 'Lifecycle' = _kuber_definitions.UNCHANGED_VALUE,
        liveness_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        name: str = _kuber_definitions.UNCHANGED_VALUE,
        ports: typing.List['ContainerPort'] = _kuber_definitions.UNCHANGED_VALUE,
        readiness_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        resources: 'ResourceRequirements' = _kuber_definitions.UNCHANGED_VALUE,
        security_context: 'SecurityContext' = _kuber_definitions.UNCHANGED_VALUE,
        startup_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        stdin: bool = _kuber_definitions.UNCHANGED_VALUE,
        stdin_once: bool = _kuber_definitions.UNCHANGED_VALUE,
        termination_message_path: str = _kuber_definitions.UNCHANGED_VALUE,
        termination_message_policy: str = _kuber_definitions.UNCHANGED_VALUE,
        tty: bool = _kuber_definitions.UNCHANGED_VALUE,
        volume_devices: typing.List['VolumeDevice'] = _kuber_definitions.UNCHANGED_VALUE,
        volume_mounts: typing.List['VolumeMount'] = _kuber_definitions.UNCHANGED_VALUE,
        working_dir: str = _kuber_definitions.UNCHANGED_VALUE,
    ) -> 'CronJobSpec':
        """Adds a container object within the specified resource."""
        values = {
            'args': args,
            'command': command,
            'env': env,
            'env_from': env_from,
            'image': image,
            'image_pull_policy': image_pull_policy,
            'lifecycle': lifecycle,
            'liveness_probe': liveness_probe,
            'name': name,
            'ports': ports,
            'readiness_probe': readiness_probe,
            'resources': resources,
            'security_context': security_context,
            'startup_probe': startup_probe,
            'stdin': stdin,
            'stdin_once': stdin_once,
            'termination_message_path': termination_message_path,
            'termination_message_policy': termination_message_policy,
            'tty': tty,
            'volume_devices': volume_devices,
            'volume_mounts': volume_mounts,
            'working_dir': working_dir,
        }
        self.job_template.spec.template.spec.containers.append(Container(**{
            k: v
            for k, v in values.items()
            if v != _kuber_definitions.UNCHANGED_VALUE
        }))
        return self

    def get_container(self, name: str) -> typing.Optional['Container']:
        """
        Fetch a container definition within this resource by name if such a
        container definition exists. Return None if no container definition
        by that name exists.
        """
        return next((c for c in self.job_template.spec.template.spec.containers if c.name == name), None)

    def get_containers(self) -> typing.List['Container']:
        """
        Returns the list of containers stored in this resource if any such
        containers exist.
        """
        return self.job_template.spec.template.spec.containers

    def __enter__(self) -> 'CronJobSpec':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CronJobStatus(_kuber_definitions.Definition):
    """
    CronJobStatus represents the current state of a cron job.
    """

    def __init__(
            self,
            active: typing.List['ObjectReference'] = None,
            last_schedule_time: str = None,
    ):
        """Create CronJobStatus instance."""
        super(CronJobStatus, self).__init__(
            api_version='batch/v1beta1',
            kind='CronJobStatus'
        )
        self._properties = {
            'active': active if active is not None else [],
            'lastScheduleTime': last_schedule_time if last_schedule_time is not None else None,

        }
        self._types = {
            'active': (list, ObjectReference),
            'lastScheduleTime': (str, None),

        }

    @property
    def active(self) -> typing.List['ObjectReference']:
        """
        A list of pointers to currently running jobs.
        """
        return self._properties.get('active')

    @active.setter
    def active(
            self,
            value: typing.Union[typing.List['ObjectReference'], typing.List[dict]]
    ):
        """
        A list of pointers to currently running jobs.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = ObjectReference().from_dict(item)
            cleaned.append(item)
        self._properties['active'] = cleaned

    @property
    def last_schedule_time(self) -> str:
        """
        Information when was the last time the job was successfully
        scheduled.
        """
        return self._properties.get('lastScheduleTime')

    @last_schedule_time.setter
    def last_schedule_time(
            self,
            value: typing.Union[str, _datetime.datetime, _datetime.date]
    ):
        """
        Information when was the last time the job was successfully
        scheduled.
        """
        if isinstance(value, _datetime.datetime):
            value = value.strftime('%Y-%m-%dT%H:%M:%SZ')
        elif isinstance(value, _datetime.date):
            value = value.strftime('%Y-%m-%dT00:00:00Z')
        self._properties['lastScheduleTime'] = value

    def __enter__(self) -> 'CronJobStatus':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class JobTemplateSpec(_kuber_definitions.Definition):
    """
    JobTemplateSpec describes the data a Job should have when
    created from a template
    """

    def __init__(
            self,
            metadata: 'ObjectMeta' = None,
            spec: 'JobSpec' = None,
    ):
        """Create JobTemplateSpec instance."""
        super(JobTemplateSpec, self).__init__(
            api_version='batch/v1beta1',
            kind='JobTemplateSpec'
        )
        self._properties = {
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'spec': spec if spec is not None else JobSpec(),

        }
        self._types = {
            'metadata': (ObjectMeta, None),
            'spec': (JobSpec, None),

        }

    @property
    def metadata(self) -> 'ObjectMeta':
        """
        Standard object's metadata of the jobs created from this
        template. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """
        Standard object's metadata of the jobs created from this
        template. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def spec(self) -> 'JobSpec':
        """
        Specification of the desired behavior of the job. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#spec-and-status
        """
        return self._properties.get('spec')

    @spec.setter
    def spec(self, value: typing.Union['JobSpec', dict]):
        """
        Specification of the desired behavior of the job. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#spec-and-status
        """
        if isinstance(value, dict):
            value = JobSpec().from_dict(value)
        self._properties['spec'] = value

    def append_container(
        self,
        args: typing.List[str] = _kuber_definitions.UNCHANGED_VALUE,
        command: typing.List[str] = _kuber_definitions.UNCHANGED_VALUE,
        env: typing.List['EnvVar'] = _kuber_definitions.UNCHANGED_VALUE,
        env_from: typing.List['EnvFromSource'] = _kuber_definitions.UNCHANGED_VALUE,
        image: str = _kuber_definitions.UNCHANGED_VALUE,
        image_pull_policy: str = _kuber_definitions.UNCHANGED_VALUE,
        lifecycle: 'Lifecycle' = _kuber_definitions.UNCHANGED_VALUE,
        liveness_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        name: str = _kuber_definitions.UNCHANGED_VALUE,
        ports: typing.List['ContainerPort'] = _kuber_definitions.UNCHANGED_VALUE,
        readiness_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        resources: 'ResourceRequirements' = _kuber_definitions.UNCHANGED_VALUE,
        security_context: 'SecurityContext' = _kuber_definitions.UNCHANGED_VALUE,
        startup_probe: 'Probe' = _kuber_definitions.UNCHANGED_VALUE,
        stdin: bool = _kuber_definitions.UNCHANGED_VALUE,
        stdin_once: bool = _kuber_definitions.UNCHANGED_VALUE,
        termination_message_path: str = _kuber_definitions.UNCHANGED_VALUE,
        termination_message_policy: str = _kuber_definitions.UNCHANGED_VALUE,
        tty: bool = _kuber_definitions.UNCHANGED_VALUE,
        volume_devices: typing.List['VolumeDevice'] = _kuber_definitions.UNCHANGED_VALUE,
        volume_mounts: typing.List['VolumeMount'] = _kuber_definitions.UNCHANGED_VALUE,
        working_dir: str = _kuber_definitions.UNCHANGED_VALUE,
    ) -> 'JobTemplateSpec':
        """Adds a container object within the specified resource."""
        values = {
            'args': args,
            'command': command,
            'env': env,
            'env_from': env_from,
            'image': image,
            'image_pull_policy': image_pull_policy,
            'lifecycle': lifecycle,
            'liveness_probe': liveness_probe,
            'name': name,
            'ports': ports,
            'readiness_probe': readiness_probe,
            'resources': resources,
            'security_context': security_context,
            'startup_probe': startup_probe,
            'stdin': stdin,
            'stdin_once': stdin_once,
            'termination_message_path': termination_message_path,
            'termination_message_policy': termination_message_policy,
            'tty': tty,
            'volume_devices': volume_devices,
            'volume_mounts': volume_mounts,
            'working_dir': working_dir,
        }
        self.spec.template.spec.containers.append(Container(**{
            k: v
            for k, v in values.items()
            if v != _kuber_definitions.UNCHANGED_VALUE
        }))
        return self

    def get_container(self, name: str) -> typing.Optional['Container']:
        """
        Fetch a container definition within this resource by name if such a
        container definition exists. Return None if no container definition
        by that name exists.
        """
        return next((c for c in self.spec.template.spec.containers if c.name == name), None)

    def get_containers(self) -> typing.List['Container']:
        """
        Returns the list of containers stored in this resource if any such
        containers exist.
        """
        return self.spec.template.spec.containers

    def __enter__(self) -> 'JobTemplateSpec':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



import os

import kuber

directory = os.path.realpath(os.path.dirname(__file__))


def test_metadata_owner_reference():
    """
    Should load the owner references correctly. Specifically that the
    apiVersion and kind are set correctly to the reference
    """
    pod = kuber.from_yaml_file(os.path.join(directory, 'owner_metadata.yaml'))
    reference = pod.metadata.owner_references[0]
    reference_dict = pod.to_dict()['metadata']['ownerReferences'][0]
    serialized = pod.to_yaml()

    assert reference.kind == 'DaemonSet'
    assert reference_dict['kind'] == 'DaemonSet'
    assert ' kind: DaemonSet' in serialized

    assert reference.api_version == 'apps/v1'
    assert reference_dict['apiVersion'] == 'apps/v1'
    assert ' apiVersion: apps/v1' in serialized



import os

import kuber

directory = os.path.realpath(os.path.dirname(__file__))


def test_odd_api_versions():
    """Should serialize and deserialize odd api versions."""
    bundle = kuber.from_file(
        os.path.join(directory, 'odd-resources.yaml'),
        kubernetes_version='1.11'
    )

    print(bundle.render_yaml_bundle())
    assert len(bundle.resources) == 1



import os

import pytest

import kuber

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))


def test_assignments():
    """Should manage internal settings as expected."""
    settings = kuber.create_bundle().settings
    settings.add(a=1, b=2)
    settings.add(b=None)
    settings.c = 'hello'
    settings['d'] = 'goodbye'

    assert settings.a == 1
    assert settings['a'] == 1
    assert settings.grab('a', 'b', 'c', 'd') == (1, None, 'hello', 'goodbye')


def test_file_error():
    """Should error out if the file is not of the correct type."""
    settings = kuber.create_bundle().settings

    with pytest.raises(IOError):
        settings.add_from_file(os.path.realpath(__file__))



import typing

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_13.meta_v1 import ListMeta
from kuber.v1_13.meta_v1 import ObjectMeta


class Initializer(_kuber_definitions.Definition):
    """
    Initializer describes the name and the failure policy of an
    initializer, and what resources it applies to.
    """

    def __init__(
            self,
            name: str = None,
            rules: typing.List['Rule'] = None,
    ):
        """Create Initializer instance."""
        super(Initializer, self).__init__(
            api_version='admissionregistration/v1alpha1',
            kind='Initializer'
        )
        self._properties = {
            'name': name if name is not None else '',
            'rules': rules if rules is not None else [],

        }
        self._types = {
            'name': (str, None),
            'rules': (list, Rule),

        }

    @property
    def name(self) -> str:
        """
        Name is the identifier of the initializer. It will be added
        to the object that needs to be initialized. Name should be
        fully qualified, e.g., alwayspullimages.kubernetes.io, where
        "alwayspullimages" is the name of the webhook, and
        kubernetes.io is the name of the organization. Required
        """
        return self._properties.get('name')

    @name.setter
    def name(self, value: str):
        """
        Name is the identifier of the initializer. It will be added
        to the object that needs to be initialized. Name should be
        fully qualified, e.g., alwayspullimages.kubernetes.io, where
        "alwayspullimages" is the name of the webhook, and
        kubernetes.io is the name of the organization. Required
        """
        self._properties['name'] = value

    @property
    def rules(self) -> typing.List['Rule']:
        """
        Rules describes what resources/subresources the initializer
        cares about. The initializer cares about an operation if it
        matches _any_ Rule. Rule.Resources must not include
        subresources.
        """
        return self._properties.get('rules')

    @rules.setter
    def rules(
            self,
            value: typing.Union[typing.List['Rule'], typing.List[dict]]
    ):
        """
        Rules describes what resources/subresources the initializer
        cares about. The initializer cares about an operation if it
        matches _any_ Rule. Rule.Resources must not include
        subresources.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = Rule().from_dict(item)
            cleaned.append(item)
        self._properties['rules'] = cleaned

    def __enter__(self) -> 'Initializer':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class InitializerConfiguration(_kuber_definitions.Resource):
    """
    InitializerConfiguration describes the configuration of
    initializers.
    """

    def __init__(
            self,
            initializers: typing.List['Initializer'] = None,
            metadata: 'ObjectMeta' = None,
    ):
        """Create InitializerConfiguration instance."""
        super(InitializerConfiguration, self).__init__(
            api_version='admissionregistration/v1alpha1',
            kind='InitializerConfiguration'
        )
        self._properties = {
            'initializers': initializers if initializers is not None else [],
            'metadata': metadata if metadata is not None else ObjectMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'initializers': (list, Initializer),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),

        }

    @property
    def initializers(self) -> typing.List['Initializer']:
        """
        Initializers is a list of resources and their default
        initializers Order-sensitive. When merging multiple
        InitializerConfigurations, we sort the initializers from
        different InitializerConfigurations by the name of the
        InitializerConfigurations; the order of the initializers
        from the same InitializerConfiguration is preserved.
        """
        return self._properties.get('initializers')

    @initializers.setter
    def initializers(
            self,
            value: typing.Union[typing.List['Initializer'], typing.List[dict]]
    ):
        """
        Initializers is a list of resources and their default
        initializers Order-sensitive. When merging multiple
        InitializerConfigurations, we sort the initializers from
        different InitializerConfigurations by the name of the
        InitializerConfigurations; the order of the initializers
        from the same InitializerConfiguration is preserved.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = Initializer().from_dict(item)
            cleaned.append(item)
        self._properties['initializers'] = cleaned

    @property
    def metadata(self) -> 'ObjectMeta':
        """
        Standard object metadata; More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata.
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """
        Standard object metadata; More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata.
        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    def create_resource(self, namespace: 'str' = None):
        """
        Creates the InitializerConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'create_namespaced_initializer_configuration',
            'create_initializer_configuration'
        ]

        _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )

    def replace_resource(self, namespace: 'str' = None):
        """
        Replaces the InitializerConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'replace_namespaced_initializer_configuration',
            'replace_initializer_configuration'
        ]

        _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def patch_resource(self, namespace: 'str' = None):
        """
        Patches the InitializerConfiguration in the currently
        configured Kubernetes cluster.
        """
        names = [
            'patch_namespaced_initializer_configuration',
            'patch_initializer_configuration'
        ]

        _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def get_resource_status(self, namespace: 'str' = None):
        """This resource does not have a status."""
        pass

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the InitializerConfiguration from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_initializer_configuration',
            'read_initializer_configuration'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the InitializerConfiguration from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_initializer_configuration',
            'delete_initializer_configuration'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.AdmissionregistrationV1alpha1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.AdmissionregistrationV1alpha1Api(**kwargs)

    def __enter__(self) -> 'InitializerConfiguration':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class InitializerConfigurationList(_kuber_definitions.Collection):
    """
    InitializerConfigurationList is a list of
    InitializerConfiguration.
    """

    def __init__(
            self,
            items: typing.List['InitializerConfiguration'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create InitializerConfigurationList instance."""
        super(InitializerConfigurationList, self).__init__(
            api_version='admissionregistration/v1alpha1',
            kind='InitializerConfigurationList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, InitializerConfiguration),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['InitializerConfiguration']:
        """
        List of InitializerConfiguration.
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['InitializerConfiguration'], typing.List[dict]]
    ):
        """
        List of InitializerConfiguration.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = InitializerConfiguration().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#types-kinds
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#types-kinds
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.AdmissionregistrationV1alpha1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.AdmissionregistrationV1alpha1Api(**kwargs)

    def __enter__(self) -> 'InitializerConfigurationList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class Rule(_kuber_definitions.Definition):
    """
    Rule is a tuple of APIGroups, APIVersion, and Resources.It
    is recommended to make sure that all the tuple expansions
    are valid.
    """

    def __init__(
            self,
            api_groups: typing.List[str] = None,
            api_versions: typing.List[str] = None,
            resources: typing.List[str] = None,
    ):
        """Create Rule instance."""
        super(Rule, self).__init__(
            api_version='admissionregistration/v1alpha1',
            kind='Rule'
        )
        self._properties = {
            'apiGroups': api_groups if api_groups is not None else [],
            'apiVersions': api_versions if api_versions is not None else [],
            'resources': resources if resources is not None else [],

        }
        self._types = {
            'apiGroups': (list, str),
            'apiVersions': (list, str),
            'resources': (list, str),

        }

    @property
    def api_groups(self) -> typing.List[str]:
        """
        APIGroups is the API groups the resources belong to. '*' is
        all groups. If '*' is present, the length of the slice must
        be one. Required.
        """
        return self._properties.get('apiGroups')

    @api_groups.setter
    def api_groups(self, value: typing.List[str]):
        """
        APIGroups is the API groups the resources belong to. '*' is
        all groups. If '*' is present, the length of the slice must
        be one. Required.
        """
        self._properties['apiGroups'] = value

    @property
    def api_versions(self) -> typing.List[str]:
        """
        APIVersions is the API versions the resources belong to. '*'
        is all versions. If '*' is present, the length of the slice
        must be one. Required.
        """
        return self._properties.get('apiVersions')

    @api_versions.setter
    def api_versions(self, value: typing.List[str]):
        """
        APIVersions is the API versions the resources belong to. '*'
        is all versions. If '*' is present, the length of the slice
        must be one. Required.
        """
        self._properties['apiVersions'] = value

    @property
    def resources(self) -> typing.List[str]:
        """
        Resources is a list of resources this rule applies to.

        For example: 'pods' means pods. 'pods/log' means the log
        subresource of pods. '*' means all resources, but not
        subresources. 'pods/*' means all subresources of pods.
        '*/scale' means all scale subresources. '*/*' means all
        resources and their subresources.

        If wildcard is present, the validation rule will ensure
        resources do not overlap with each other.

        Depending on the enclosing object, subresources might not be
        allowed. Required.
        """
        return self._properties.get('resources')

    @resources.setter
    def resources(self, value: typing.List[str]):
        """
        Resources is a list of resources this rule applies to.

        For example: 'pods' means pods. 'pods/log' means the log
        subresource of pods. '*' means all resources, but not
        subresources. 'pods/*' means all subresources of pods.
        '*/scale' means all scale subresources. '*/*' means all
        resources and their subresources.

        If wildcard is present, the validation rule will ensure
        resources do not overlap with each other.

        Depending on the enclosing object, subresources might not be
        allowed. Required.
        """
        self._properties['resources'] = value

    def __enter__(self) -> 'Rule':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

from kuber import kube_api
from kuber.latest import core_v1


def test_execute():
    """"Should execute the specified action on the service without error."""
    service = core_v1.Service()
    service.metadata.name = 'foo'

    api = MagicMock()
    code = MagicMock()
    code.co_varnames = ('bob', 'foo', 'namespace')
    api.create_namespaced_service.__code__ = code
    with patch.object(service, 'get_resource_api') as get_api:
        get_api.return_value = api
        service.create_resource(namespace='foo')

    api.create_namespaced_service.assert_called_once()


def test_execute_no_name():
    """"Should raise error when no api client function name exists."""
    resource = MagicMock()
    resource.get_resource_api.return_value = None

    with pytest.raises(ValueError):
        kube_api.execute(
            action='foo',
            resource=resource,
            names=['spam', 'ham']
        )



import kuber
import pathlib

from pytest import mark

RESOURCES_PATH = pathlib.Path(__file__).parent.joinpath('resources.yaml')

SCENARIOS = (
    {
        'filters': ['first/*'],
        'matches': set(),
    },
    {
        'filters': [],
        'matches': {
            'first/Pod/foo-pod-first',
            'first/Deployment/foo-deployment-first',
            'second/Pod/foo-pod-second',
            'second/Job/foo-job-second',
            'second/Pod/bar-pod-second',
            'second/Deployment/bar-deployment-second',
            'second/Job/bar-job-second',
        },
    },
    {
        'filters': ['*-job-*', '*-deployment-first'],
        'matches': {
            'first/Deployment/foo-deployment-first',
            'second/Job/foo-job-second',
            'second/Job/bar-job-second',
        },
    },
    {
        'filters': ['foo-*'],
        'matches': {
            'first/Pod/foo-pod-first',
            'first/Deployment/foo-deployment-first',
            'second/Pod/foo-pod-second',
            'second/Job/foo-job-second',
        },
    },
    {
        'filters': ['FIRST/*/*'],
        'matches': {
            'first/Pod/foo-pod-first',
            'first/Deployment/foo-deployment-first',
        },
    },
)


@mark.parametrize('scenario', SCENARIOS)
def test_array_filtering(scenario: dict):
    """Should filter down to the expected resources."""
    bundle = kuber.from_file(str(RESOURCES_PATH))
    matches = bundle.resources.matching(*scenario['filters'])

    assert scenario['matches'] == {
        f'{m.metadata.namespace}/{m.kind}/{m.metadata.name}'
        for m in matches.to_list()
    }



from pytest import mark

from kuber import kube_api

SCENARIOS = (
    ({'foo': None, 'fooBar': ''}, {'fooBar': ''}),
    ({'foo': False, 'foo_bar': 42}, {'foo': False, 'fooBar': 42}),
    ({'foo_bar_baz': 'a'}, {'fooBarBaz': 'a'}),
)


@mark.parametrize('source, expected', SCENARIOS)
def test_to_kuber_dict(source: dict, expected: dict):
    """Should convert to the expected value."""
    assert kube_api.to_kuber_dict(source) == expected



import typing

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_14.meta_v1 import ListMeta
from kuber.v1_14.meta_v1 import ObjectMeta


class RuntimeClass(_kuber_definitions.Resource):
    """
    RuntimeClass defines a class of container runtime supported
    in the cluster. The RuntimeClass is used to determine which
    container runtime is used to run all containers in a pod.
    RuntimeClasses are (currently) manually defined by a user or
    cluster provisioner, and referenced in the PodSpec. The
    Kubelet is responsible for resolving the RuntimeClassName
    reference before running the pod.  For more details, see
    https://git.k8s.io/enhancements/keps/sig-node/runtime-
    class.md
    """

    def __init__(
            self,
            handler: str = None,
            metadata: 'ObjectMeta' = None,
    ):
        """Create RuntimeClass instance."""
        super(RuntimeClass, self).__init__(
            api_version='node/v1beta1',
            kind='RuntimeClass'
        )
        self._properties = {
            'handler': handler if handler is not None else '',
            'metadata': metadata if metadata is not None else ObjectMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'handler': (str, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),

        }

    @property
    def handler(self) -> str:
        """
        Handler specifies the underlying runtime and configuration
        that the CRI implementation will use to handle pods of this
        class. The possible values are specific to the node & CRI
        configuration.  It is assumed that all handlers are
        available on every node, and handlers of the same name are
        equivalent on every node. For example, a handler called
        "runc" might specify that the runc OCI runtime (using native
        Linux containers) will be used to run the containers in a
        pod. The Handler must conform to the DNS Label (RFC 1123)
        requirements, and is immutable.
        """
        return self._properties.get('handler')

    @handler.setter
    def handler(self, value: str):
        """
        Handler specifies the underlying runtime and configuration
        that the CRI implementation will use to handle pods of this
        class. The possible values are specific to the node & CRI
        configuration.  It is assumed that all handlers are
        available on every node, and handlers of the same name are
        equivalent on every node. For example, a handler called
        "runc" might specify that the runc OCI runtime (using native
        Linux containers) will be used to run the containers in a
        pod. The Handler must conform to the DNS Label (RFC 1123)
        requirements, and is immutable.
        """
        self._properties['handler'] = value

    @property
    def metadata(self) -> 'ObjectMeta':
        """
        More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """
        More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    def create_resource(self, namespace: 'str' = None):
        """
        Creates the RuntimeClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'create_namespaced_runtime_class',
            'create_runtime_class'
        ]

        _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )

    def replace_resource(self, namespace: 'str' = None):
        """
        Replaces the RuntimeClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'replace_namespaced_runtime_class',
            'replace_runtime_class'
        ]

        _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def patch_resource(self, namespace: 'str' = None):
        """
        Patches the RuntimeClass in the currently
        configured Kubernetes cluster.
        """
        names = [
            'patch_namespaced_runtime_class',
            'patch_runtime_class'
        ]

        _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def get_resource_status(self, namespace: 'str' = None):
        """This resource does not have a status."""
        pass

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the RuntimeClass from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_runtime_class',
            'read_runtime_class'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the RuntimeClass from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_runtime_class',
            'delete_runtime_class'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.NodeV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.NodeV1beta1Api(**kwargs)

    def __enter__(self) -> 'RuntimeClass':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class RuntimeClassList(_kuber_definitions.Collection):
    """
    RuntimeClassList is a list of RuntimeClass objects.
    """

    def __init__(
            self,
            items: typing.List['RuntimeClass'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create RuntimeClassList instance."""
        super(RuntimeClassList, self).__init__(
            api_version='node/v1beta1',
            kind='RuntimeClassList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, RuntimeClass),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['RuntimeClass']:
        """
        Items is a list of schema objects.
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['RuntimeClass'], typing.List[dict]]
    ):
        """
        Items is a list of schema objects.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = RuntimeClass().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/api-
        conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.NodeV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.NodeV1beta1Api(**kwargs)

    def __enter__(self) -> 'RuntimeClassList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False
